_D='region'
_C='size'
_B=False
_A=None
import inspect,json,logging,os,re,shutil,zipfile
from enum import Enum
from typing import Any,Dict,List,Optional,Set,Tuple,Type,Union
from deepdiff import DeepDiff
from localstack.utils.common import json_safe,merge_recursive,mkdir,new_tmp_dir,rm_rf,to_str
from localstack.utils.generic.singleton_utils import SubtypesInstanceManager
from localstack_ext.bootstrap.pods.constants import COMPRESSION_FORMAT,METAMODELS_FILE,VERSION_SERVICE_INFO_FILE
from localstack_ext.bootstrap.pods.models import StateFileRef,Version
from localstack_ext.bootstrap.pods.object_storage import ObjectStorageProvider,get_object_storage_provider
from localstack_ext.bootstrap.pods.utils.common import PodsConfigContext,add_file_to_archive,read_file_from_archive
from localstack_ext.bootstrap.pods.utils.hash_utils import compute_file_hash,random_hash
from localstack_ext.bootstrap.state_utils import API_STATES_DIR,api_states_traverse,check_already_visited,get_object_dict,load_persisted_object
LOG=logging.getLogger(__name__)
CAPTURE_METAMODEL_SEPARATELY=_B
PLACEHOLDER_NO_CHANGE={'_meta_':'no-change'}
class MetamodelDeltaMethod(Enum):SIMPLE='simple';DEEP_DIFF='deepdiff'
class MetamodelDelta(SubtypesInstanceManager):
	def create_delta_for_state_files(G,sf1,sf2,config_context):
		E=config_context;B=sf2;A=sf1
		if A and B:
			if A.hash_ref==B.hash_ref:return PLACEHOLDER_NO_CHANGE
			else:C=load_persisted_object(E.get_obj_file_path(A.hash_ref));D=load_persisted_object(E.get_obj_file_path(B.hash_ref));return G.create_delta_json(C,D)
		if not A:D=load_persisted_object(E.get_obj_file_path(B.hash_ref));F=type(D);C=_infer_backend_init(F,B)
		else:C=load_persisted_object(E.get_obj_file_path(A.hash_ref));F=type(C);D=_infer_backend_init(F,A)
		H=G.create_delta_json(C,D);return H
	def create_delta_log(E,state_from,state_to,config_context):
		D=config_context;H=_filter_special_cases(state_from);O=H[0];P=H[1];Q=H[2];A=_filter_special_cases(state_to);L=A[0];R=A[1];S=A[2]
		def T(state_files):return{os.path.join(A.rel_path,A.file_name):A for A in state_files}
		C={};M=T(L)
		for B in O:
			I=C.setdefault(B.region,{});J=I.setdefault(B.service,{})
			if B.any_congruence(L):F=os.path.join(B.rel_path,B.file_name);U=M.pop(F);G=E.create_delta_for_state_files(B,U,D)
			else:G=E.create_delta_for_state_files(B,_A,D)
			J[B.file_name]=G
		for A in M.values():I=C.setdefault(A.region,{});J=I.setdefault(A.service,{});G=E.create_delta_for_state_files(_A,A,D);J[A.file_name]=G
		def N(service,sf_from,sf_to):
			P='name';K=service;L={};F={};M=get_object_storage_provider(D)
			for N in sf_from:
				A=N.region;G=L.setdefault(A,{});B=load_persisted_object(M.get_state_file_location_by_key(N.hash_ref))
				if hasattr(B,P):G[B.name]=B
			for O in sf_to:
				A=O.region;G=F.setdefault(A,{});B=load_persisted_object(M.get_state_file_location_by_key(O.hash_ref))
				if hasattr(B,P):G[B.name]=B
			for (A,Q) in L.items():H=F.pop(A,{});I=C.setdefault(A,{});J=E.create_delta_json(Q,H);I[K]=J
			for (A,H) in F.items():I=C.setdefault(A,{});J=E.create_delta_json({},H);I[K]=J
		N('sqs',Q,S);N('s3',P,R);K=os.path.join(D.get_delta_log_path(),random_hash());C=json_safe(C)
		with open(K,'w')as V:json.dump(C,V,indent=1)
		F=compute_file_hash(K);W=os.path.join(D.get_delta_log_path(),F);os.rename(K,W);return F
	def create_delta_json(A,state1,state2):raise NotImplementedError
class MetamodelDeltaDeepDiff(MetamodelDelta):
	@staticmethod
	def impl_name():return MetamodelDeltaMethod.DEEP_DIFF
	def create_delta_json(A,state1,state2):return DeepDiff(state1,state2).to_json()
class MetamodelDeltaSimple(MetamodelDelta):
	@staticmethod
	def impl_name():return MetamodelDeltaMethod.SIMPLE
	def create_delta_json(C,state1,state2):
		B=_create_metamodel_helper(state1)or{};A=_create_metamodel_helper(state2)or{}
		if B==A:return PLACEHOLDER_NO_CHANGE
		return A
class CommitMetamodelUtils:
	def __init__(A,config_context,object_storage=_A):
		B=object_storage;from localstack_ext.bootstrap.pods.object_storage import get_object_storage_provider as C;A.config_context=config_context;A.object_storage=B
		if B is _A:A.object_storage=C(A.config_context)
	def create_metadata_archive(A,version,delete_reference=_B,overwrite=_B,metamodels_file=_A):
		I=overwrite;E=metamodels_file;D=version;C=A.object_storage.get_revision_or_version_by_key(D.active_revision_ptr if I else D.incoming_revision_ptr);L=C.revision_number;B=A.config_context.metadata_dir(D.version_number);mkdir(B)
		if E:
			J=A.config_context.get_version_meta_archive_path(D.version_number)
			if os.path.isfile(J):
				with zipfile.ZipFile(J)as M:M.extractall(B)
		while C:
			F=C.assoc_commit
			if not F:break
			K=F.delta_log_ptr
			if K:
				G=A.object_storage.get_delta_file_by_key(K,A.config_context.get_delta_log_path())
				if not G:continue
				N=A.config_context.commit_metamodel_file(C.revision_number+1);O=os.path.join(A.config_context.pod_root_dir,B,N);shutil.copy(G,O)
				if delete_reference:os.remove(G)
			P=F.head_ptr if I else C.parent_ptr;C=A.object_storage.get_revision_by_key(P)
		Q,R=A.create_delta_metamodel_file(D.version_number,L)
		if CAPTURE_METAMODEL_SEPARATELY:
			E=E or METAMODELS_FILE;S=os.path.join(B,E)
			with open(S,'w')as H:json.dump(Q,H)
		T=os.path.join(B,VERSION_SERVICE_INFO_FILE)
		with open(T,'w')as H:json.dump(R,H,indent=1)
		shutil.make_archive(B,COMPRESSION_FORMAT,root_dir=B);rm_rf(B)
	def create_metamodel_from_state_files(E,version):
		A=new_tmp_dir();B=E.config_context.get_version_state_archive(version=version)
		if not B:return
		with zipfile.ZipFile(B)as F:F.extractall(A)
		G=os.path.join(A,API_STATES_DIR);C={};D={};api_states_traverse(api_states_path=G,side_effect=_metadata_create_func,mutables=[C,D]);rm_rf(A);return C,D
	@classmethod
	def get_metamodel_delta(L,prev_metamodel,this_metamodel):
		C=prev_metamodel;A=this_metamodel
		if not C:return A
		def H(prev_service_state,service_state):return service_state!=prev_service_state
		D={};A=A or{}
		for (E,I) in A.items():
			D[E]=F={};J=C.get(E)or{}
			for (B,G) in I.items():
				F[B]=PLACEHOLDER_NO_CHANGE;K=J.get(B)
				if H(K,G):F[B]=G
		return D
	def create_delta_metamodel_file(A,version,revision,store_to_zip=_B):
		F=store_to_zip;C=revision;B=version;D,G=A.create_metamodel_from_state_files(version=B);H=A.config_context.get_version_meta_archive_path(version=B);I=A.config_context.metamodel_file(revision=C)
		if C<=1:
			if F:add_file_to_archive(H,entry_name=I,content=json.dumps(D))
			return D,G
		E=_A
		if CAPTURE_METAMODEL_SEPARATELY:
			J=A.reconstruct_metamodel(version=B,revision=C-1);E=A.get_metamodel_delta(J,D)
			if F:add_file_to_archive(H,entry_name=I,content=json.dumps(E))
		return E,G
	def reconstruct_metamodel(D,version,revision=_A):
		E=version;B=revision
		if not B:B=infer_max_commit_diff(archive_path=D.config_context.get_version_meta_archive(E))
		A={}
		for I in range(1,B+1):
			F=D.get_version_metamodel(version=E,revision=I)
			if not F:return{}
			for (C,G) in F.items():
				if C not in A:A[C]=G;continue
				for (J,H) in G.items():
					if H==PLACEHOLDER_NO_CHANGE:continue
					A[C][J]=H
		return A
	def get_version_metamodel(A,version,revision=_A):
		B=A.config_context.get_version_meta_archive(version)
		if B:C=A.config_context.metamodel_file(revision=revision);D=read_file_from_archive(B,C);return json.loads(D)
	def get_commit_diff(B,version_no,commit_no):
		C=version_no;D=B.config_context.get_version_meta_archive(C)
		if not D:LOG.warning(f"No metadata found for version {C}");return
		E=B.config_context.commit_metamodel_file(commit_no);A=read_file_from_archive(archive_path=D,file_name=E);A=json.loads(to_str(A or'{}'));return A
def _infer_backend_init(clazz,sf):
	A=clazz
	if isinstance(A,dict):return{}
	D=getattr(A,'__init__',_A);C=inspect.getfullargspec(D)
	if _D in C.args:B=A(region=sf.region)
	elif'region_name'in C.args:B=A(region_name=sf.region)
	else:B=A()
	return B
def _filter_special_cases(state_files):
	B,C,D=[],[],[]
	for A in state_files:
		if A.service=='sqs':D.append(A)
		elif A.service=='s3':C.append(A)
		else:B.append(A)
	return B,C,D
def _metadata_create_func(**A):
	try:J=A.get('dir_name');B=A.get('account_id');K=A.get('fname');E=A.get(_D);F=A.get('service_name');G=A.get('mutables');C=G[0];C[B]=C.get(B)or{};H=os.path.join(J,K);L=load_persisted_object(H);M=_create_metamodel_helper(L)or{};I=C.get(B)[E]=C.get(B).get(E)or{};N=I[F]=I.get(F)or{};D=G[1];merge_recursive(source=M,destination=N);O=D.setdefault(E,{});D=O.setdefault(F,{});D[_C]=D.get(_C,0)+os.path.getsize(H)
	except Exception as P:LOG.exception(f"Unable to create metamodel for state object {A} : {P}")
def _create_metamodel_helper(obj,width=25,visited=_A):
	D=visited;C=width;A=obj
	if A is _A:return A
	F,D=check_already_visited(A,D)
	if F:return A
	E=get_object_dict(A);B=A=E if E is not _A else A
	if isinstance(A,dict):
		B=dict(B)
		for (G,H) in B.items():B[G]=_create_metamodel_helper(H,width=C,visited=D)
	elif isinstance(A,list):
		B=[_create_metamodel_helper(B,width=C,visited=D)for B in A]
		if len(B)>C:B={_C:len(B),'items':B[:C]}
	return B
def infer_max_commit_diff(archive_path):
	A=1
	with zipfile.ZipFile(archive_path)as D:
		for E in D.namelist():
			B=re.search('metamodel_commit_([0-9]+)\\.json',E)
			if not B:continue
			C=int(B.group(1))
			if C>A:A=C
	return A


_D='command'
_C='root:ssh_proxy'
_B='result'
_A='op'
import json,logging,os,time
from typing import Dict,List
import requests
from localstack import config
from localstack.aws.api import RequestContext
from localstack.aws.api.ec2 import ImportImageRequest,ImportImageResult,Reservation,RunInstancesRequest
from localstack.config import in_docker
from localstack.constants import LOCALHOST
from localstack.services.generic_proxy import RegionBackend
from localstack.utils.common import get_free_tcp_port,to_str
from localstack_ext.config import DEFAULT_PORT_LOCAL_DAEMON
from localstack_ext.services.ec2.vmmanager.base import InternalError,VmManager
LOG=logging.getLogger(__name__)
VBOX_COMMAND='VBoxManage'
VBOX_PREFIX='localstack-ec2'
VM_SEED_ISO_URL='https://github.com/localstack/localstack-artifacts/raw/master/ec2-iso/seed.iso'
HOST_OS=None
class UnixVirtualBox:
	def create_vm(G,image,instance,disk):M='host';I=instance;H=image;G.check_vm_env();LOG.info('Start VM instance "%s" from image "%s"',I,H);C=VBOX_COMMAND;J=f"{C} storageattach";E={_A:'s3:download','file_name':'s3file.%s'%H,'bucket':disk['s3Bucket'],'key':disk['s3Key'],'overwrite':False};D=call_local_daemon(E);K=D['local_file'];A=f"{C} internalcommands sethduuid {K}";run_local_daemon_cmd(A);B=f"{VBOX_PREFIX}{I}";A=f"{C} createvm --name {B} --register";run_local_daemon_cmd(A);A=f'{C} storagectl {B} --name "SATA" --add sata --bootable on --portcount 1';run_local_daemon_cmd(A);A=f"{J} {B} --storagectl SATA --port 0 --type hdd --medium {K}";run_local_daemon_cmd(A);F=os.path.join(config.dirs.functions,'vm_seed.iso');A='test -e %s || wget -O %s %s'%(F,F,VM_SEED_ISO_URL);run_local_daemon_cmd(A);A=f"{J} {B} --storagectl SATA --port 1 --type dvddrive --medium {F}";run_local_daemon_cmd(A);E={_A:_C};D=call_local_daemon(E);L=D.get('forward_port')or get_free_tcp_port();N=D.get(M);A=f'{C} modifyvm {B} --memory 1024 --nic1 nat --natpf1 "ssh,tcp,,{L},,22"';run_local_daemon_cmd(A);G.pre_startup(B);A=f"{C} startvm {B} --type headless";D=run_local_daemon_cmd(A);time.sleep(3);return{_B:D,M:N,'ssh_port':L}
	def pre_startup(A,vm_name):0
	def remove_vm(B,instance_id):A=f"{VBOX_COMMAND} controlvm MyVM poweroff";print(A)
	def check_vm_env(C):
		A={_A:'shell',_D:'which %s'%VBOX_COMMAND};B=call_local_daemon(A)
		if not B.get(_B):raise InternalError('Please install VirtualBox and VBoxManage utility on the host system')
	def cleanup(D):
		try:
			A=run_local_daemon_cmd('%s list vms'%VBOX_COMMAND);A=A.split('\n')
			for C in A:
				B=C.split(' ')[0].strip('"')
				if B.startswith(VBOX_PREFIX):run_local_daemon_cmd(f"{VBOX_COMMAND} unregistervm {B} --delete")
		except Exception:pass
class MacOsVirtualBox(UnixVirtualBox):0
class LinuxVirtualBox(UnixVirtualBox):
	def pre_startup(B,vm_name):A=f"{VBOX_COMMAND} modifyvm {vm_name} --nictype1 virtio";run_local_daemon_cmd(A)
def get_virtualbox():
	global HOST_OS
	if not HOST_OS:A=call_local_daemon({_A:'getos'});HOST_OS=A[_B];LOG.debug('Determined host operating system type: %s',HOST_OS)
	if HOST_OS=='macos':return MacOsVirtualBox()
	if HOST_OS=='linux':return LinuxVirtualBox()
	raise InternalError('Host operating system not supported: %s'%HOST_OS)
def run_local_daemon_cmd(cmd):
	B={_A:'shell',_D:cmd};A=call_local_daemon(B)
	if _B in A:return A[_B]
	raise InternalError('Error running command: %s'%A.get('error'))
def call_local_daemon(data):
	B=config.DOCKER_HOST_FROM_CONTAINER if in_docker()else LOCALHOST;C=f"http://{B}:{DEFAULT_PORT_LOCAL_DAEMON}";A=requests.post(C,json.dumps(data))
	if A.status_code>=400:raise InternalError('Error calling : %s'%A.content)
	return json.loads(to_str(A.content))
def is_daemon_running():
	A={_A:_C}
	try:call_local_daemon(A);return True
	except Exception:return False
class Ec2Backend(RegionBackend):
	disk_containers:Dict[str,List[Dict]];instance_hosts:Dict[str,str]
	def __init__(A):super().__init__();A.disk_containers={};A.instance_hosts={}
class VirtualBoxVmManager(VmManager):
	@staticmethod
	def impl_name():return'virtualbox'
	def import_image(A,context,import_image_request):raise NotImplementedError
	def run_instances(A,context,run_instances_request):raise NotImplementedError


_O='_deployed'
_N='cognito-identity'
_M='PoolName'
_L='ClientName'
_K='IdentityPoolId'
_J='delete'
_I='Domain'
_H='GroupName'
_G='cognito-idp'
_F='parameters'
_E='create'
_D='ProviderName'
_C=None
_B='UserPoolId'
_A='function'
from localstack.constants import TEST_AWS_ACCOUNT_ID,TRUE_STRINGS
from localstack.services.cloudformation.deployment_utils import PLACEHOLDER_RESOURCE_NAME
from localstack.services.cloudformation.service_models import REF_ATTRS,REF_ID_ATTRS,GenericBaseModel
from localstack.utils.aws import aws_stack
from localstack_ext.services.cognito_idp.cognito_utils import get_issuer_url
from localstack_ext.utils.aws import aws_utils
class CognitoUserPool(GenericBaseModel):
	@staticmethod
	def cloudformation_type():return'AWS::Cognito::UserPool'
	def get_cfn_attribute(B,attribute_name):
		A=attribute_name
		if A in REF_ATTRS:return B.get_physical_resource_id(A)
		C=B._get_id()
		if A=='Arn':return aws_utils.cognito_userpool_arn(C)
		if A==_D:return 'cognito-idp.{r}.amazonaws.com/{r}_{a}'.format(r=aws_stack.get_region(),a=TEST_AWS_ACCOUNT_ID)
		if A=='ProviderURL':return get_issuer_url(pool_id=C)
		return super(CognitoUserPool,B).get_cfn_attribute(A)
	def get_physical_resource_id(B,attribute,**C):
		A=B._get_id()
		if not A:return A
		if attribute in REF_ATTRS:return A
		return aws_utils.cognito_userpool_arn(A)
	def get_resource_name(A):return A.props.get(_M)
	def _get_id(B):A=B.props;return A.get(_B)or A.get('Id')
	@classmethod
	def fetch_details(C,pool_name):A=aws_stack.connect_to_service(_G);B=A.list_user_pools(MaxResults=100)['UserPools'];return([A for A in B if A['Name']==pool_name]or[_C])[0]
	@staticmethod
	def get_deploy_templates():
		def A(params,**Q):
			K='MaxLength';J='MinLength';I='Schema';H='Policies';F='MinimumLength';D='AutoVerifiedAttributes';L=[H,'LambdaConfig',D,'AliasAttributes','UsernameAttributes','VerificationMessageTemplate','EmailVerificationMessage','EmailVerificationSubject','SmsVerificationMessage','SmsAuthenticationMessage','DeviceConfiguration','EmailConfiguration','SmsConfiguration','UserPoolTags','AdminCreateUserConfig',I,'UserPoolAddOns'];G={A:A for A in L};G[_M]='UserPoolName';A={A:params.get(B)for(A,B)in G.items()};M=A.get(H)or{};B=M.get('PasswordPolicy')or{}
			if B.get(F):B[F]=int(B[F])
			N=['RequireLowercase','RequireNumbers','RequireSymbols','RequireUppercase'];O=list(TRUE_STRINGS)+[True]
			for E in N:
				if E in B:B[E]=B[E]in O
			if isinstance(A.get(D),str):A[D]=[A[D]]
			for P in A.get(I)or[]:
				C=P.get('StringAttributeConstraints')
				if C:C[J]=str(C.get(J,1));C[K]=str(C.get(K,50))
			return A
		return{_E:{_A:'create_user_pool',_F:A}}
class UserPoolGroup(GenericBaseModel):
	@staticmethod
	def cloudformation_type():return'AWS::Cognito::UserPoolGroup'
	def get_physical_resource_id(A,attribute,**B):return A.props.get(_H)
	def fetch_state(A,stack_name,resources):C=resources;B=stack_name;E=aws_stack.connect_to_service(_G);D=A.props;F=A.resolve_refs_recursively(B,D.get(_H),C);G=A.resolve_refs_recursively(B,D.get(_B),C);H=E.list_groups(UserPoolId=G)['Groups'];I=[A for A in H if A[_H]==F];return(I or[_C])[0]
	@staticmethod
	def get_deploy_templates():return{_E:{_A:'create_group'},_J:{_A:'delete_group',_F:[_H,_B]}}
class IdentityPool(GenericBaseModel):
	@staticmethod
	def cloudformation_type():return'AWS::Cognito::IdentityPool'
	def get_cfn_attribute(B,attribute_name):
		A=attribute_name
		try:return super(IdentityPool,B).get_cfn_attribute(A)
		except Exception:
			if A in REF_ID_ATTRS:return B.get_physical_resource_id(A)
			if A==_D:return 'cognito.{r}.amazonaws.com/{r}_{a}'.format(r=aws_stack.get_region(),a=TEST_AWS_ACCOUNT_ID)
			raise
	def get_physical_resource_id(A,attribute,**B):
		if attribute in REF_ID_ATTRS:return A.props.get(_K)
	@classmethod
	def fetch_details(D,pool_name):A=aws_stack.connect_to_service(_N);B=A.list_identity_pools(MaxResults=100)['IdentityPools'];C=[A for A in B if A['IdentityPoolName']==pool_name];return(C or[_C])[0]
	@staticmethod
	def get_deploy_templates():return{_E:{_A:'create_identity_pool'}}
class CognitoUserPoolClient(GenericBaseModel):
	@staticmethod
	def cloudformation_type():return'AWS::Cognito::UserPoolClient'
	def get_cfn_attribute(B,attribute_name):
		A=attribute_name
		if A in REF_ID_ATTRS:return B.get_physical_resource_id(A)
		return super(CognitoUserPoolClient,B).get_cfn_attribute(A)
	def get_physical_resource_id(A,attribute,**B):
		if attribute in REF_ID_ATTRS:return A.props.get('ClientId')
	@classmethod
	def fetch_details(C,pool_id,client_name):A=aws_stack.connect_to_service(_G);B=A.list_user_pool_clients(UserPoolId=pool_id)['UserPoolClients'];return([A for A in B if A[_L]==client_name]or[_C])[0]
	@staticmethod
	def get_deploy_templates():
		def A(params,**C):
			B='RefreshTokenValidity';A=params
			if not A.get(_L):A[_L]=PLACEHOLDER_RESOURCE_NAME
			if A.get(B)is not _C:A[B]=int(A[B])
			return A
		return{_E:{_A:'create_user_pool_client',_F:A}}
class CognitoUserPoolDomain(GenericBaseModel):
	@staticmethod
	def cloudformation_type():return'AWS::Cognito::UserPoolDomain'
	def get_physical_resource_id(A,attribute,**B):
		if attribute in REF_ID_ATTRS:return A.props.get(_I)
	def fetch_state(A,stack_name,resources):B=aws_stack.connect_to_service(_G);C=A.resolve_refs_recursively(stack_name,A.props.get(_I),resources);D=B.describe_user_pool_domain(Domain=C)['DomainDescription'];return D or _C
	@staticmethod
	def get_deploy_templates():return{_E:{_A:'create_user_pool_domain'},_J:{_A:'delete_user_pool_domain',_F:{_B:_B,_I:_I}}}
class CognitoIdentityPoolRoleAttachment(GenericBaseModel):
	@staticmethod
	def cloudformation_type():return'AWS::Cognito::IdentityPoolRoleAttachment'
	def fetch_state(B,stack_name,resources):
		C=aws_stack.connect_to_service(_N);D=B.resolve_refs_recursively(stack_name,B.props.get(_K),resources);A=C.get_identity_pool_roles(IdentityPoolId=D)
		if A:A[_O]=True
		return A or _C
	def get_physical_resource_id(B,attribute,**C):A=B.props;return A.get(_O)and'cognito-pool-roles-%s'%A.get(_K)
	@staticmethod
	def get_deploy_templates():
		def A(params,*H,**I):
			F='IdentityProvider';E='RoleMappings';B=params;D=B.get(E)
			if D:
				C={}
				for (G,A) in D.items():
					if A.get(F):C[A.pop(F)]=A
					else:C[G]=A
				B[E]=C
			return B
		return{_E:{_A:'set_identity_pool_roles',_F:A}}
class CognitoUserPoolIdentityProvider(GenericBaseModel):
	@staticmethod
	def cloudformation_type():return'AWS::Cognito::UserPoolIdentityProvider'
	def get_physical_resource_id(A,attribute,**B):return A.props.get(_D)
	def fetch_state(A,stack_name,resources):C=resources;B=stack_name;E=aws_stack.connect_to_service(_G);D=A.props;F=A.resolve_refs_recursively(B,D.get(_B),C);G=A.resolve_refs_recursively(B,D.get(_D),C);H=E.list_identity_providers(UserPoolId=F)['Providers'];I=[A for A in H if A[_D]==G];return(I or[_C])[0]
	@staticmethod
	def get_deploy_templates():return{_E:{_A:'create_identity_provider'},_J:{_A:'delete_identity_provider',_F:{_B:_B,_D:_D}}}


_A='[red]Error:[/red] '
import logging,os,sys
from typing import Any
import click
from localstack.cli import LocalstackCli,LocalstackCliPlugin,console
from localstack.utils.analytics.cli import publish_invocation
from .cloud_pods import pod
class ProCliPlugin(LocalstackCliPlugin):
	name='pro'
	def should_load(B):A=os.getenv('LOCALSTACK_API_KEY');return True if A else False
	def is_active(A):return A.should_load()
	def attach(B,cli):A=cli.group;A.add_command(cmd_login);A.add_command(cmd_logout);A.add_command(daemons);A.add_command(dns);A.add_command(pod)
@click.group(name='daemons',help='Manage local daemon processes')
def daemons():0
@click.command(name='login',help='Log in with your account credentials')
@click.option('--username',help='Username for login')
@publish_invocation
def cmd_login(username):
	from localstack_ext.bootstrap import auth
	try:auth.login(username);console.print('successfully logged in')
	except Exception as A:console.print('authentication error: %s'%A)
@click.command(name='logout',help='Log out and delete any session tokens')
@publish_invocation
def cmd_logout():
	from localstack_ext.bootstrap import auth
	try:auth.logout();console.print('successfully logged out')
	except Exception as A:console.print('logout error: %s'%A)
@daemons.command(name='start',help='Start local daemon processes')
@publish_invocation
def cmd_daemons_start():from localstack_ext.bootstrap import local_daemon as A;console.log('Starting local daemons processes ...');B=A.start_in_background();B.join()
@daemons.command(name='stop',help='Stop local daemon processes')
@publish_invocation
def cmd_daemons_stop():from localstack_ext.bootstrap import local_daemon as A;console.log('Stopping local daemons processes ...');A.kill_servers()
@daemons.command(name='log',help='Show log of daemon process')
@publish_invocation
def cmd_daemons_log():
	from localstack_ext.bootstrap import local_daemon as B;A=B.get_log_file_path()
	if not os.path.isfile(A):console.print('no log found')
	else:
		with open(A,'r')as C:
			for D in C:sys.stdout.write(D);sys.stdout.flush()
@click.group(name='dns',help='Manage DNS settings of your host')
def dns():0
@dns.command(name='systemd-resolved',help='Manage DNS settings of systemd-resolved (Ubuntu, Debian etc.)')
@click.option('--revert',is_flag=True,help='Revert systemd-resolved settings for the docker interface')
@publish_invocation
def cmd_dns_systemd(revert):import localstack_ext.bootstrap.dns_utils;from localstack_ext.bootstrap.dns_utils import configure_systemd as A;console.print('Configuring systemd-resolved...');B=localstack_ext.bootstrap.dns_utils.LOG.name;localstack_ext.bootstrap.dns_utils.LOG=ConsoleLogger(B);A(revert)
class ConsoleLogger(logging.Logger):
	def __init__(A,name):super(ConsoleLogger,A).__init__(name)
	def info(B,msg,*A,**C):console.print(msg%A)
	def warning(B,msg,*A,**C):console.print('[red]Warning:[/red] ',msg%A)
	def error(B,msg,*A,**C):console.print(_A,msg%A)
	def exception(B,msg,*A,**C):console.print(_A,msg%A);console.print_exception()


import sys
from datetime import datetime
from typing import Dict, List, Optional

if sys.version_info >= (3, 8):
    from typing import TypedDict
else:
    from typing_extensions import TypedDict

from localstack.aws.api import RequestContext, ServiceException, ServiceRequest, handler

AmazonResourceName = str
Boolean = bool
BoxedBoolean = bool
CatalogNameString = str
CommentString = str
DatabaseString = str
DescriptionString = str
ErrorCategory = int
ErrorCode = str
ErrorMessage = str
ErrorType = int
ExpressionString = str
IdempotencyToken = str
Integer = int
KeyString = str
MaxDataCatalogsCount = int
MaxDatabasesCount = int
MaxEngineVersionsCount = int
MaxNamedQueriesCount = int
MaxPreparedStatementsCount = int
MaxQueryExecutionsCount = int
MaxQueryResults = int
MaxTableMetadataCount = int
MaxTagsCount = int
MaxWorkGroupsCount = int
NameString = str
NamedQueryDescriptionString = str
NamedQueryId = str
ParametersMapValue = str
QueryExecutionId = str
QueryString = str
StatementName = str
String = str
TableTypeString = str
TagKey = str
TagValue = str
Token = str
TypeString = str
WorkGroupDescriptionString = str
WorkGroupName = str
datumString = str


class ColumnNullable(str):
    NOT_NULL = "NOT_NULL"
    NULLABLE = "NULLABLE"
    UNKNOWN = "UNKNOWN"


class DataCatalogType(str):
    LAMBDA = "LAMBDA"
    GLUE = "GLUE"
    HIVE = "HIVE"


class EncryptionOption(str):
    SSE_S3 = "SSE_S3"
    SSE_KMS = "SSE_KMS"
    CSE_KMS = "CSE_KMS"


class QueryExecutionState(str):
    QUEUED = "QUEUED"
    RUNNING = "RUNNING"
    SUCCEEDED = "SUCCEEDED"
    FAILED = "FAILED"
    CANCELLED = "CANCELLED"


class S3AclOption(str):
    BUCKET_OWNER_FULL_CONTROL = "BUCKET_OWNER_FULL_CONTROL"


class StatementType(str):
    DDL = "DDL"
    DML = "DML"
    UTILITY = "UTILITY"


class ThrottleReason(str):
    CONCURRENT_QUERY_LIMIT_EXCEEDED = "CONCURRENT_QUERY_LIMIT_EXCEEDED"


class WorkGroupState(str):
    ENABLED = "ENABLED"
    DISABLED = "DISABLED"


class InternalServerException(ServiceException):
    """Indicates a platform issue, which may be due to a transient condition or
    outage.
    """

    Message: Optional[ErrorMessage]


class InvalidRequestException(ServiceException):
    """Indicates that something is wrong with the input to the request. For
    example, a required parameter may be missing or out of range.
    """

    AthenaErrorCode: Optional[ErrorCode]
    Message: Optional[ErrorMessage]


class MetadataException(ServiceException):
    """An exception that Athena received when it called a custom metastore.
    Occurs if the error is not caused by user input
    (``InvalidRequestException``) or from the Athena platform
    (``InternalServerException``). For example, if a user-created Lambda
    function is missing permissions, the Lambda ``4XX`` exception is
    returned in a ``MetadataException``.
    """

    Message: Optional[ErrorMessage]


class ResourceNotFoundException(ServiceException):
    """A resource, such as a workgroup, was not found."""

    Message: Optional[ErrorMessage]
    ResourceName: Optional[AmazonResourceName]


class TooManyRequestsException(ServiceException):
    """Indicates that the request was throttled."""

    Message: Optional[ErrorMessage]
    Reason: Optional[ThrottleReason]


class AclConfiguration(TypedDict, total=False):
    """Indicates that an Amazon S3 canned ACL should be set to control
    ownership of stored query results. When Athena stores query results in
    Amazon S3, the canned ACL is set with the ``x-amz-acl`` request header.
    For more information about S3 Object Ownership, see `Object Ownership
    settings <https://docs.aws.amazon.com/AmazonS3/latest/userguide/about-object-ownership.html#object-ownership-overview>`__
    in the *Amazon S3 User Guide*.
    """

    S3AclOption: S3AclOption


class AthenaError(TypedDict, total=False):
    """Provides information about an Athena query error. The ``AthenaError``
    feature provides standardized error information to help you understand
    failed queries and take steps after a query failure occurs.
    ``AthenaError`` includes an ``ErrorCategory`` field that specifies
    whether the cause of the failed query is due to system error, user
    error, or other error.
    """

    ErrorCategory: Optional[ErrorCategory]
    ErrorType: Optional[ErrorType]
    Retryable: Optional[Boolean]
    ErrorMessage: Optional[String]


NamedQueryIdList = List[NamedQueryId]


class BatchGetNamedQueryInput(ServiceRequest):
    NamedQueryIds: NamedQueryIdList


class UnprocessedNamedQueryId(TypedDict, total=False):
    """Information about a named query ID that could not be processed."""

    NamedQueryId: Optional[NamedQueryId]
    ErrorCode: Optional[ErrorCode]
    ErrorMessage: Optional[ErrorMessage]


UnprocessedNamedQueryIdList = List[UnprocessedNamedQueryId]


class NamedQuery(TypedDict, total=False):
    """A query, where ``QueryString`` contains the SQL statements that make up
    the query.
    """

    Name: NameString
    Description: Optional[DescriptionString]
    Database: DatabaseString
    QueryString: QueryString
    NamedQueryId: Optional[NamedQueryId]
    WorkGroup: Optional[WorkGroupName]


NamedQueryList = List[NamedQuery]


class BatchGetNamedQueryOutput(TypedDict, total=False):
    NamedQueries: Optional[NamedQueryList]
    UnprocessedNamedQueryIds: Optional[UnprocessedNamedQueryIdList]


QueryExecutionIdList = List[QueryExecutionId]


class BatchGetQueryExecutionInput(ServiceRequest):
    QueryExecutionIds: QueryExecutionIdList


class UnprocessedQueryExecutionId(TypedDict, total=False):
    """Describes a query execution that failed to process."""

    QueryExecutionId: Optional[QueryExecutionId]
    ErrorCode: Optional[ErrorCode]
    ErrorMessage: Optional[ErrorMessage]


UnprocessedQueryExecutionIdList = List[UnprocessedQueryExecutionId]


class EngineVersion(TypedDict, total=False):
    """The Athena engine version for running queries."""

    SelectedEngineVersion: Optional[NameString]
    EffectiveEngineVersion: Optional[NameString]


Long = int


class QueryExecutionStatistics(TypedDict, total=False):
    """The amount of data scanned during the query execution and the amount of
    time that it took to execute, and the type of statement that was run.
    """

    EngineExecutionTimeInMillis: Optional[Long]
    DataScannedInBytes: Optional[Long]
    DataManifestLocation: Optional[String]
    TotalExecutionTimeInMillis: Optional[Long]
    QueryQueueTimeInMillis: Optional[Long]
    QueryPlanningTimeInMillis: Optional[Long]
    ServiceProcessingTimeInMillis: Optional[Long]


Date = datetime


class QueryExecutionStatus(TypedDict, total=False):
    """The completion date, current state, submission time, and state change
    reason (if applicable) for the query execution.
    """

    State: Optional[QueryExecutionState]
    StateChangeReason: Optional[String]
    SubmissionDateTime: Optional[Date]
    CompletionDateTime: Optional[Date]
    AthenaError: Optional[AthenaError]


class QueryExecutionContext(TypedDict, total=False):
    """The database and data catalog context in which the query execution
    occurs.
    """

    Database: Optional[DatabaseString]
    Catalog: Optional[CatalogNameString]


class EncryptionConfiguration(TypedDict, total=False):
    """If query results are encrypted in Amazon S3, indicates the encryption
    option used (for example, ``SSE_KMS`` or ``CSE_KMS``) and key
    information.
    """

    EncryptionOption: EncryptionOption
    KmsKey: Optional[String]


class ResultConfiguration(TypedDict, total=False):
    """The location in Amazon S3 where query results are stored and the
    encryption option, if any, used for query results. These are known as
    "client-side settings". If workgroup settings override client-side
    settings, then the query uses the workgroup settings.
    """

    OutputLocation: Optional[String]
    EncryptionConfiguration: Optional[EncryptionConfiguration]
    ExpectedBucketOwner: Optional[String]
    AclConfiguration: Optional[AclConfiguration]


class QueryExecution(TypedDict, total=False):
    """Information about a single instance of a query execution."""

    QueryExecutionId: Optional[QueryExecutionId]
    Query: Optional[QueryString]
    StatementType: Optional[StatementType]
    ResultConfiguration: Optional[ResultConfiguration]
    QueryExecutionContext: Optional[QueryExecutionContext]
    Status: Optional[QueryExecutionStatus]
    Statistics: Optional[QueryExecutionStatistics]
    WorkGroup: Optional[WorkGroupName]
    EngineVersion: Optional[EngineVersion]


QueryExecutionList = List[QueryExecution]


class BatchGetQueryExecutionOutput(TypedDict, total=False):
    QueryExecutions: Optional[QueryExecutionList]
    UnprocessedQueryExecutionIds: Optional[UnprocessedQueryExecutionIdList]


BytesScannedCutoffValue = int


class Column(TypedDict, total=False):
    """Contains metadata for a column in a table."""

    Name: NameString
    Type: Optional[TypeString]
    Comment: Optional[CommentString]


class ColumnInfo(TypedDict, total=False):
    """Information about the columns in a query execution result."""

    CatalogName: Optional[String]
    SchemaName: Optional[String]
    TableName: Optional[String]
    Name: String
    Label: Optional[String]
    Type: String
    Precision: Optional[Integer]
    Scale: Optional[Integer]
    Nullable: Optional[ColumnNullable]
    CaseSensitive: Optional[Boolean]


ColumnInfoList = List[ColumnInfo]
ColumnList = List[Column]


class Tag(TypedDict, total=False):
    """A label that you assign to a resource. In Athena, a resource can be a
    workgroup or data catalog. Each tag consists of a key and an optional
    value, both of which you define. For example, you can use tags to
    categorize Athena workgroups or data catalogs by purpose, owner, or
    environment. Use a consistent set of tag keys to make it easier to
    search and filter workgroups or data catalogs in your account. For best
    practices, see `Tagging Best
    Practices <https://aws.amazon.com/answers/account-management/aws-tagging-strategies/>`__.
    Tag keys can be from 1 to 128 UTF-8 Unicode characters, and tag values
    can be from 0 to 256 UTF-8 Unicode characters. Tags can use letters and
    numbers representable in UTF-8, and the following characters: + - = . _
    : / @. Tag keys and values are case-sensitive. Tag keys must be unique
    per resource. If you specify more than one tag, separate them by commas.
    """

    Key: Optional[TagKey]
    Value: Optional[TagValue]


TagList = List[Tag]
ParametersMap = Dict[KeyString, ParametersMapValue]


class CreateDataCatalogInput(ServiceRequest):
    Name: CatalogNameString
    Type: DataCatalogType
    Description: Optional[DescriptionString]
    Parameters: Optional[ParametersMap]
    Tags: Optional[TagList]


class CreateDataCatalogOutput(TypedDict, total=False):
    pass


class CreateNamedQueryInput(ServiceRequest):
    Name: NameString
    Description: Optional[DescriptionString]
    Database: DatabaseString
    QueryString: QueryString
    ClientRequestToken: Optional[IdempotencyToken]
    WorkGroup: Optional[WorkGroupName]


class CreateNamedQueryOutput(TypedDict, total=False):
    NamedQueryId: Optional[NamedQueryId]


class CreatePreparedStatementInput(ServiceRequest):
    StatementName: StatementName
    WorkGroup: WorkGroupName
    QueryStatement: QueryString
    Description: Optional[DescriptionString]


class CreatePreparedStatementOutput(TypedDict, total=False):
    pass


class WorkGroupConfiguration(TypedDict, total=False):
    """The configuration of the workgroup, which includes the location in
    Amazon S3 where query results are stored, the encryption option, if any,
    used for query results, whether the Amazon CloudWatch Metrics are
    enabled for the workgroup and whether workgroup settings override query
    settings, and the data usage limits for the amount of data scanned per
    query or per workgroup. The workgroup settings override is specified in
    ``EnforceWorkGroupConfiguration`` (true/false) in the
    ``WorkGroupConfiguration``. See
    WorkGroupConfiguration$EnforceWorkGroupConfiguration.
    """

    ResultConfiguration: Optional[ResultConfiguration]
    EnforceWorkGroupConfiguration: Optional[BoxedBoolean]
    PublishCloudWatchMetricsEnabled: Optional[BoxedBoolean]
    BytesScannedCutoffPerQuery: Optional[BytesScannedCutoffValue]
    RequesterPaysEnabled: Optional[BoxedBoolean]
    EngineVersion: Optional[EngineVersion]


class CreateWorkGroupInput(ServiceRequest):
    Name: WorkGroupName
    Configuration: Optional[WorkGroupConfiguration]
    Description: Optional[WorkGroupDescriptionString]
    Tags: Optional[TagList]


class CreateWorkGroupOutput(TypedDict, total=False):
    pass


class DataCatalog(TypedDict, total=False):
    """Contains information about a data catalog in an Amazon Web Services
    account.
    """

    Name: CatalogNameString
    Description: Optional[DescriptionString]
    Type: DataCatalogType
    Parameters: Optional[ParametersMap]


class DataCatalogSummary(TypedDict, total=False):
    """The summary information for the data catalog, which includes its name
    and type.
    """

    CatalogName: Optional[CatalogNameString]
    Type: Optional[DataCatalogType]


DataCatalogSummaryList = List[DataCatalogSummary]


class Database(TypedDict, total=False):
    """Contains metadata information for a database in a data catalog."""

    Name: NameString
    Description: Optional[DescriptionString]
    Parameters: Optional[ParametersMap]


DatabaseList = List[Database]


class Datum(TypedDict, total=False):
    """A piece of data (a field in the table)."""

    VarCharValue: Optional[datumString]


class DeleteDataCatalogInput(ServiceRequest):
    Name: CatalogNameString


class DeleteDataCatalogOutput(TypedDict, total=False):
    pass


class DeleteNamedQueryInput(ServiceRequest):
    NamedQueryId: NamedQueryId


class DeleteNamedQueryOutput(TypedDict, total=False):
    pass


class DeletePreparedStatementInput(ServiceRequest):
    StatementName: StatementName
    WorkGroup: WorkGroupName


class DeletePreparedStatementOutput(TypedDict, total=False):
    pass


class DeleteWorkGroupInput(ServiceRequest):
    WorkGroup: WorkGroupName
    RecursiveDeleteOption: Optional[BoxedBoolean]


class DeleteWorkGroupOutput(TypedDict, total=False):
    pass


EngineVersionsList = List[EngineVersion]


class GetDataCatalogInput(ServiceRequest):
    Name: CatalogNameString


class GetDataCatalogOutput(TypedDict, total=False):
    DataCatalog: Optional[DataCatalog]


class GetDatabaseInput(ServiceRequest):
    CatalogName: CatalogNameString
    DatabaseName: NameString


class GetDatabaseOutput(TypedDict, total=False):
    Database: Optional[Database]


class GetNamedQueryInput(ServiceRequest):
    NamedQueryId: NamedQueryId


class GetNamedQueryOutput(TypedDict, total=False):
    NamedQuery: Optional[NamedQuery]


class GetPreparedStatementInput(ServiceRequest):
    StatementName: StatementName
    WorkGroup: WorkGroupName


class PreparedStatement(TypedDict, total=False):
    """A prepared SQL statement for use with Athena."""

    StatementName: Optional[StatementName]
    QueryStatement: Optional[QueryString]
    WorkGroupName: Optional[WorkGroupName]
    Description: Optional[DescriptionString]
    LastModifiedTime: Optional[Date]


class GetPreparedStatementOutput(TypedDict, total=False):
    PreparedStatement: Optional[PreparedStatement]


class GetQueryExecutionInput(ServiceRequest):
    QueryExecutionId: QueryExecutionId


class GetQueryExecutionOutput(TypedDict, total=False):
    QueryExecution: Optional[QueryExecution]


class GetQueryResultsInput(ServiceRequest):
    QueryExecutionId: QueryExecutionId
    NextToken: Optional[Token]
    MaxResults: Optional[MaxQueryResults]


class ResultSetMetadata(TypedDict, total=False):
    """The metadata that describes the column structure and data types of a
    table of query results. To return a ``ResultSetMetadata`` object, use
    GetQueryResults.
    """

    ColumnInfo: Optional[ColumnInfoList]


datumList = List[Datum]


class Row(TypedDict, total=False):
    """The rows that make up a query result table."""

    Data: Optional[datumList]


RowList = List[Row]


class ResultSet(TypedDict, total=False):
    """The metadata and rows that make up a query result set. The metadata
    describes the column structure and data types. To return a ``ResultSet``
    object, use GetQueryResults.
    """

    Rows: Optional[RowList]
    ResultSetMetadata: Optional[ResultSetMetadata]


class GetQueryResultsOutput(TypedDict, total=False):
    UpdateCount: Optional[Long]
    ResultSet: Optional[ResultSet]
    NextToken: Optional[Token]


class GetTableMetadataInput(ServiceRequest):
    CatalogName: CatalogNameString
    DatabaseName: NameString
    TableName: NameString


Timestamp = datetime


class TableMetadata(TypedDict, total=False):
    """Contains metadata for a table."""

    Name: NameString
    CreateTime: Optional[Timestamp]
    LastAccessTime: Optional[Timestamp]
    TableType: Optional[TableTypeString]
    Columns: Optional[ColumnList]
    PartitionKeys: Optional[ColumnList]
    Parameters: Optional[ParametersMap]


class GetTableMetadataOutput(TypedDict, total=False):
    TableMetadata: Optional[TableMetadata]


class GetWorkGroupInput(ServiceRequest):
    WorkGroup: WorkGroupName


class WorkGroup(TypedDict, total=False):
    """A workgroup, which contains a name, description, creation time, state,
    and other configuration, listed under WorkGroup$Configuration. Each
    workgroup enables you to isolate queries for you or your group of users
    from other queries in the same account, to configure the query results
    location and the encryption configuration (known as workgroup settings),
    to enable sending query metrics to Amazon CloudWatch, and to establish
    per-query data usage control limits for all queries in a workgroup. The
    workgroup settings override is specified in
    ``EnforceWorkGroupConfiguration`` (true/false) in the
    ``WorkGroupConfiguration``. See
    WorkGroupConfiguration$EnforceWorkGroupConfiguration.
    """

    Name: WorkGroupName
    State: Optional[WorkGroupState]
    Configuration: Optional[WorkGroupConfiguration]
    Description: Optional[WorkGroupDescriptionString]
    CreationTime: Optional[Date]


class GetWorkGroupOutput(TypedDict, total=False):
    WorkGroup: Optional[WorkGroup]


class ListDataCatalogsInput(ServiceRequest):
    NextToken: Optional[Token]
    MaxResults: Optional[MaxDataCatalogsCount]


class ListDataCatalogsOutput(TypedDict, total=False):
    DataCatalogsSummary: Optional[DataCatalogSummaryList]
    NextToken: Optional[Token]


class ListDatabasesInput(ServiceRequest):
    CatalogName: CatalogNameString
    NextToken: Optional[Token]
    MaxResults: Optional[MaxDatabasesCount]


class ListDatabasesOutput(TypedDict, total=False):
    DatabaseList: Optional[DatabaseList]
    NextToken: Optional[Token]


class ListEngineVersionsInput(ServiceRequest):
    NextToken: Optional[Token]
    MaxResults: Optional[MaxEngineVersionsCount]


class ListEngineVersionsOutput(TypedDict, total=False):
    EngineVersions: Optional[EngineVersionsList]
    NextToken: Optional[Token]


class ListNamedQueriesInput(ServiceRequest):
    NextToken: Optional[Token]
    MaxResults: Optional[MaxNamedQueriesCount]
    WorkGroup: Optional[WorkGroupName]


class ListNamedQueriesOutput(TypedDict, total=False):
    NamedQueryIds: Optional[NamedQueryIdList]
    NextToken: Optional[Token]


class ListPreparedStatementsInput(ServiceRequest):
    WorkGroup: WorkGroupName
    NextToken: Optional[Token]
    MaxResults: Optional[MaxPreparedStatementsCount]


class PreparedStatementSummary(TypedDict, total=False):
    """The name and last modified time of the prepared statement."""

    StatementName: Optional[StatementName]
    LastModifiedTime: Optional[Date]


PreparedStatementsList = List[PreparedStatementSummary]


class ListPreparedStatementsOutput(TypedDict, total=False):
    PreparedStatements: Optional[PreparedStatementsList]
    NextToken: Optional[Token]


class ListQueryExecutionsInput(ServiceRequest):
    NextToken: Optional[Token]
    MaxResults: Optional[MaxQueryExecutionsCount]
    WorkGroup: Optional[WorkGroupName]


class ListQueryExecutionsOutput(TypedDict, total=False):
    QueryExecutionIds: Optional[QueryExecutionIdList]
    NextToken: Optional[Token]


class ListTableMetadataInput(ServiceRequest):
    CatalogName: CatalogNameString
    DatabaseName: NameString
    Expression: Optional[ExpressionString]
    NextToken: Optional[Token]
    MaxResults: Optional[MaxTableMetadataCount]


TableMetadataList = List[TableMetadata]


class ListTableMetadataOutput(TypedDict, total=False):
    TableMetadataList: Optional[TableMetadataList]
    NextToken: Optional[Token]


class ListTagsForResourceInput(ServiceRequest):
    ResourceARN: AmazonResourceName
    NextToken: Optional[Token]
    MaxResults: Optional[MaxTagsCount]


class ListTagsForResourceOutput(TypedDict, total=False):
    Tags: Optional[TagList]
    NextToken: Optional[Token]


class ListWorkGroupsInput(ServiceRequest):
    NextToken: Optional[Token]
    MaxResults: Optional[MaxWorkGroupsCount]


class WorkGroupSummary(TypedDict, total=False):
    """The summary information for the workgroup, which includes its name,
    state, description, and the date and time it was created.
    """

    Name: Optional[WorkGroupName]
    State: Optional[WorkGroupState]
    Description: Optional[WorkGroupDescriptionString]
    CreationTime: Optional[Date]
    EngineVersion: Optional[EngineVersion]


WorkGroupsList = List[WorkGroupSummary]


class ListWorkGroupsOutput(TypedDict, total=False):
    WorkGroups: Optional[WorkGroupsList]
    NextToken: Optional[Token]


class ResultConfigurationUpdates(TypedDict, total=False):
    """The information about the updates in the query results, such as output
    location and encryption configuration for the query results.
    """

    OutputLocation: Optional[String]
    RemoveOutputLocation: Optional[BoxedBoolean]
    EncryptionConfiguration: Optional[EncryptionConfiguration]
    RemoveEncryptionConfiguration: Optional[BoxedBoolean]
    ExpectedBucketOwner: Optional[String]
    RemoveExpectedBucketOwner: Optional[BoxedBoolean]
    AclConfiguration: Optional[AclConfiguration]
    RemoveAclConfiguration: Optional[BoxedBoolean]


class StartQueryExecutionInput(ServiceRequest):
    QueryString: QueryString
    ClientRequestToken: Optional[IdempotencyToken]
    QueryExecutionContext: Optional[QueryExecutionContext]
    ResultConfiguration: Optional[ResultConfiguration]
    WorkGroup: Optional[WorkGroupName]


class StartQueryExecutionOutput(TypedDict, total=False):
    QueryExecutionId: Optional[QueryExecutionId]


class StopQueryExecutionInput(ServiceRequest):
    QueryExecutionId: QueryExecutionId


class StopQueryExecutionOutput(TypedDict, total=False):
    pass


TagKeyList = List[TagKey]


class TagResourceInput(ServiceRequest):
    ResourceARN: AmazonResourceName
    Tags: TagList


class TagResourceOutput(TypedDict, total=False):
    pass


class UntagResourceInput(ServiceRequest):
    ResourceARN: AmazonResourceName
    TagKeys: TagKeyList


class UntagResourceOutput(TypedDict, total=False):
    pass


class UpdateDataCatalogInput(ServiceRequest):
    Name: CatalogNameString
    Type: DataCatalogType
    Description: Optional[DescriptionString]
    Parameters: Optional[ParametersMap]


class UpdateDataCatalogOutput(TypedDict, total=False):
    pass


class UpdateNamedQueryInput(ServiceRequest):
    NamedQueryId: NamedQueryId
    Name: NameString
    Description: Optional[NamedQueryDescriptionString]
    QueryString: QueryString


class UpdateNamedQueryOutput(TypedDict, total=False):
    pass


class UpdatePreparedStatementInput(ServiceRequest):
    StatementName: StatementName
    WorkGroup: WorkGroupName
    QueryStatement: QueryString
    Description: Optional[DescriptionString]


class UpdatePreparedStatementOutput(TypedDict, total=False):
    pass


class WorkGroupConfigurationUpdates(TypedDict, total=False):
    """The configuration information that will be updated for this workgroup,
    which includes the location in Amazon S3 where query results are stored,
    the encryption option, if any, used for query results, whether the
    Amazon CloudWatch Metrics are enabled for the workgroup, whether the
    workgroup settings override the client-side settings, and the data usage
    limit for the amount of bytes scanned per query, if it is specified.
    """

    EnforceWorkGroupConfiguration: Optional[BoxedBoolean]
    ResultConfigurationUpdates: Optional[ResultConfigurationUpdates]
    PublishCloudWatchMetricsEnabled: Optional[BoxedBoolean]
    BytesScannedCutoffPerQuery: Optional[BytesScannedCutoffValue]
    RemoveBytesScannedCutoffPerQuery: Optional[BoxedBoolean]
    RequesterPaysEnabled: Optional[BoxedBoolean]
    EngineVersion: Optional[EngineVersion]


class UpdateWorkGroupInput(ServiceRequest):
    WorkGroup: WorkGroupName
    Description: Optional[WorkGroupDescriptionString]
    ConfigurationUpdates: Optional[WorkGroupConfigurationUpdates]
    State: Optional[WorkGroupState]


class UpdateWorkGroupOutput(TypedDict, total=False):
    pass


class AthenaApi:

    service = "athena"
    version = "2017-05-18"

    @handler("BatchGetNamedQuery")
    def batch_get_named_query(
        self, context: RequestContext, named_query_ids: NamedQueryIdList
    ) -> BatchGetNamedQueryOutput:
        """Returns the details of a single named query or a list of up to 50
        queries, which you provide as an array of query ID strings. Requires you
        to have access to the workgroup in which the queries were saved. Use
        ListNamedQueriesInput to get the list of named query IDs in the
        specified workgroup. If information could not be retrieved for a
        submitted query ID, information about the query ID submitted is listed
        under UnprocessedNamedQueryId. Named queries differ from executed
        queries. Use BatchGetQueryExecutionInput to get details about each
        unique query execution, and ListQueryExecutionsInput to get a list of
        query execution IDs.

        :param named_query_ids: An array of query IDs.
        :returns: BatchGetNamedQueryOutput
        :raises InternalServerException:
        :raises InvalidRequestException:
        """
        raise NotImplementedError

    @handler("BatchGetQueryExecution")
    def batch_get_query_execution(
        self, context: RequestContext, query_execution_ids: QueryExecutionIdList
    ) -> BatchGetQueryExecutionOutput:
        """Returns the details of a single query execution or a list of up to 50
        query executions, which you provide as an array of query execution ID
        strings. Requires you to have access to the workgroup in which the
        queries ran. To get a list of query execution IDs, use
        ListQueryExecutionsInput$WorkGroup. Query executions differ from named
        (saved) queries. Use BatchGetNamedQueryInput to get details about named
        queries.

        :param query_execution_ids: An array of query execution IDs.
        :returns: BatchGetQueryExecutionOutput
        :raises InternalServerException:
        :raises InvalidRequestException:
        """
        raise NotImplementedError

    @handler("CreateDataCatalog", expand=False)
    def create_data_catalog(
        self, context: RequestContext, request: CreateDataCatalogInput
    ) -> CreateDataCatalogOutput:
        """Creates (registers) a data catalog with the specified name and
        properties. Catalogs created are visible to all users of the same Amazon
        Web Services account.

        :param name: The name of the data catalog to create.
        :param type: The type of data catalog to create: ``LAMBDA`` for a federated catalog,
        ``HIVE`` for an external hive metastore, or ``GLUE`` for an Glue Data
        Catalog.
        :param description: A description of the data catalog to be created.
        :param parameters: Specifies the Lambda function or functions to use for creating the data
        catalog.
        :param tags: A list of comma separated tags to add to the data catalog that is
        created.
        :returns: CreateDataCatalogOutput
        :raises InternalServerException:
        :raises InvalidRequestException:
        """
        raise NotImplementedError

    @handler("CreateNamedQuery")
    def create_named_query(
        self,
        context: RequestContext,
        name: NameString,
        database: DatabaseString,
        query_string: QueryString,
        description: DescriptionString = None,
        client_request_token: IdempotencyToken = None,
        work_group: WorkGroupName = None,
    ) -> CreateNamedQueryOutput:
        """Creates a named query in the specified workgroup. Requires that you have
        access to the workgroup.

        For code samples using the Amazon Web Services SDK for Java, see
        `Examples and Code
        Samples <http://docs.aws.amazon.com/athena/latest/ug/code-samples.html>`__
        in the *Amazon Athena User Guide*.

        :param name: The query name.
        :param database: The database to which the query belongs.
        :param query_string: The contents of the query with all query statements.
        :param description: The query description.
        :param client_request_token: A unique case-sensitive string used to ensure the request to create the
        query is idempotent (executes only once).
        :param work_group: The name of the workgroup in which the named query is being created.
        :returns: CreateNamedQueryOutput
        :raises InternalServerException:
        :raises InvalidRequestException:
        """
        raise NotImplementedError

    @handler("CreatePreparedStatement")
    def create_prepared_statement(
        self,
        context: RequestContext,
        statement_name: StatementName,
        work_group: WorkGroupName,
        query_statement: QueryString,
        description: DescriptionString = None,
    ) -> CreatePreparedStatementOutput:
        """Creates a prepared statement for use with SQL queries in Athena.

        :param statement_name: The name of the prepared statement.
        :param work_group: The name of the workgroup to which the prepared statement belongs.
        :param query_statement: The query string for the prepared statement.
        :param description: The description of the prepared statement.
        :returns: CreatePreparedStatementOutput
        :raises InternalServerException:
        :raises InvalidRequestException:
        """
        raise NotImplementedError

    @handler("CreateWorkGroup")
    def create_work_group(
        self,
        context: RequestContext,
        name: WorkGroupName,
        configuration: WorkGroupConfiguration = None,
        description: WorkGroupDescriptionString = None,
        tags: TagList = None,
    ) -> CreateWorkGroupOutput:
        """Creates a workgroup with the specified name.

        :param name: The workgroup name.
        :param configuration: The configuration for the workgroup, which includes the location in
        Amazon S3 where query results are stored, the encryption configuration,
        if any, used for encrypting query results, whether the Amazon CloudWatch
        Metrics are enabled for the workgroup, the limit for the amount of bytes
        scanned (cutoff) per query, if it is specified, and whether workgroup's
        settings (specified with ``EnforceWorkGroupConfiguration``) in the
        ``WorkGroupConfiguration`` override client-side settings.
        :param description: The workgroup description.
        :param tags: A list of comma separated tags to add to the workgroup that is created.
        :returns: CreateWorkGroupOutput
        :raises InternalServerException:
        :raises InvalidRequestException:
        """
        raise NotImplementedError

    @handler("DeleteDataCatalog")
    def delete_data_catalog(
        self, context: RequestContext, name: CatalogNameString
    ) -> DeleteDataCatalogOutput:
        """Deletes a data catalog.

        :param name: The name of the data catalog to delete.
        :returns: DeleteDataCatalogOutput
        :raises InternalServerException:
        :raises InvalidRequestException:
        """
        raise NotImplementedError

    @handler("DeleteNamedQuery")
    def delete_named_query(
        self, context: RequestContext, named_query_id: NamedQueryId
    ) -> DeleteNamedQueryOutput:
        """Deletes the named query if you have access to the workgroup in which the
        query was saved.

        For code samples using the Amazon Web Services SDK for Java, see
        `Examples and Code
        Samples <http://docs.aws.amazon.com/athena/latest/ug/code-samples.html>`__
        in the *Amazon Athena User Guide*.

        :param named_query_id: The unique ID of the query to delete.
        :returns: DeleteNamedQueryOutput
        :raises InternalServerException:
        :raises InvalidRequestException:
        """
        raise NotImplementedError

    @handler("DeletePreparedStatement")
    def delete_prepared_statement(
        self, context: RequestContext, statement_name: StatementName, work_group: WorkGroupName
    ) -> DeletePreparedStatementOutput:
        """Deletes the prepared statement with the specified name from the
        specified workgroup.

        :param statement_name: The name of the prepared statement to delete.
        :param work_group: The workgroup to which the statement to be deleted belongs.
        :returns: DeletePreparedStatementOutput
        :raises InternalServerException:
        :raises InvalidRequestException:
        :raises ResourceNotFoundException:
        """
        raise NotImplementedError

    @handler("DeleteWorkGroup")
    def delete_work_group(
        self,
        context: RequestContext,
        work_group: WorkGroupName,
        recursive_delete_option: BoxedBoolean = None,
    ) -> DeleteWorkGroupOutput:
        """Deletes the workgroup with the specified name. The primary workgroup
        cannot be deleted.

        :param work_group: The unique name of the workgroup to delete.
        :param recursive_delete_option: The option to delete the workgroup and its contents even if the
        workgroup contains any named queries or query executions.
        :returns: DeleteWorkGroupOutput
        :raises InternalServerException:
        :raises InvalidRequestException:
        """
        raise NotImplementedError

    @handler("GetDataCatalog")
    def get_data_catalog(
        self, context: RequestContext, name: CatalogNameString
    ) -> GetDataCatalogOutput:
        """Returns the specified data catalog.

        :param name: The name of the data catalog to return.
        :returns: GetDataCatalogOutput
        :raises InternalServerException:
        :raises InvalidRequestException:
        """
        raise NotImplementedError

    @handler("GetDatabase")
    def get_database(
        self, context: RequestContext, catalog_name: CatalogNameString, database_name: NameString
    ) -> GetDatabaseOutput:
        """Returns a database object for the specified database and data catalog.

        :param catalog_name: The name of the data catalog that contains the database to return.
        :param database_name: The name of the database to return.
        :returns: GetDatabaseOutput
        :raises InternalServerException:
        :raises InvalidRequestException:
        :raises MetadataException:
        """
        raise NotImplementedError

    @handler("GetNamedQuery")
    def get_named_query(
        self, context: RequestContext, named_query_id: NamedQueryId
    ) -> GetNamedQueryOutput:
        """Returns information about a single query. Requires that you have access
        to the workgroup in which the query was saved.

        :param named_query_id: The unique ID of the query.
        :returns: GetNamedQueryOutput
        :raises InternalServerException:
        :raises InvalidRequestException:
        """
        raise NotImplementedError

    @handler("GetPreparedStatement")
    def get_prepared_statement(
        self, context: RequestContext, statement_name: StatementName, work_group: WorkGroupName
    ) -> GetPreparedStatementOutput:
        """Retrieves the prepared statement with the specified name from the
        specified workgroup.

        :param statement_name: The name of the prepared statement to retrieve.
        :param work_group: The workgroup to which the statement to be retrieved belongs.
        :returns: GetPreparedStatementOutput
        :raises InternalServerException:
        :raises InvalidRequestException:
        :raises ResourceNotFoundException:
        """
        raise NotImplementedError

    @handler("GetQueryExecution")
    def get_query_execution(
        self, context: RequestContext, query_execution_id: QueryExecutionId
    ) -> GetQueryExecutionOutput:
        """Returns information about a single execution of a query if you have
        access to the workgroup in which the query ran. Each time a query
        executes, information about the query execution is saved with a unique
        ID.

        :param query_execution_id: The unique ID of the query execution.
        :returns: GetQueryExecutionOutput
        :raises InternalServerException:
        :raises InvalidRequestException:
        """
        raise NotImplementedError

    @handler("GetQueryResults")
    def get_query_results(
        self,
        context: RequestContext,
        query_execution_id: QueryExecutionId,
        next_token: Token = None,
        max_results: MaxQueryResults = None,
    ) -> GetQueryResultsOutput:
        """Streams the results of a single query execution specified by
        ``QueryExecutionId`` from the Athena query results location in Amazon
        S3. For more information, see `Query
        Results <https://docs.aws.amazon.com/athena/latest/ug/querying.html>`__
        in the *Amazon Athena User Guide*. This request does not execute the
        query but returns results. Use StartQueryExecution to run a query.

        To stream query results successfully, the IAM principal with permission
        to call ``GetQueryResults`` also must have permissions to the Amazon S3
        ``GetObject`` action for the Athena query results location.

        IAM principals with permission to the Amazon S3 ``GetObject`` action for
        the query results location are able to retrieve query results from
        Amazon S3 even if permission to the ``GetQueryResults`` action is
        denied. To restrict user or role access, ensure that Amazon S3
        permissions to the Athena query location are denied.

        :param query_execution_id: The unique ID of the query execution.
        :param next_token: A token generated by the Athena service that specifies where to continue
        pagination if a previous request was truncated.
        :param max_results: The maximum number of results (rows) to return in this request.
        :returns: GetQueryResultsOutput
        :raises InternalServerException:
        :raises InvalidRequestException:
        """
        raise NotImplementedError

    @handler("GetTableMetadata")
    def get_table_metadata(
        self,
        context: RequestContext,
        catalog_name: CatalogNameString,
        database_name: NameString,
        table_name: NameString,
    ) -> GetTableMetadataOutput:
        """Returns table metadata for the specified catalog, database, and table.

        :param catalog_name: The name of the data catalog that contains the database and table
        metadata to return.
        :param database_name: The name of the database that contains the table metadata to return.
        :param table_name: The name of the table for which metadata is returned.
        :returns: GetTableMetadataOutput
        :raises InternalServerException:
        :raises InvalidRequestException:
        :raises MetadataException:
        """
        raise NotImplementedError

    @handler("GetWorkGroup")
    def get_work_group(
        self, context: RequestContext, work_group: WorkGroupName
    ) -> GetWorkGroupOutput:
        """Returns information about the workgroup with the specified name.

        :param work_group: The name of the workgroup.
        :returns: GetWorkGroupOutput
        :raises InternalServerException:
        :raises InvalidRequestException:
        """
        raise NotImplementedError

    @handler("ListDataCatalogs")
    def list_data_catalogs(
        self,
        context: RequestContext,
        next_token: Token = None,
        max_results: MaxDataCatalogsCount = None,
    ) -> ListDataCatalogsOutput:
        """Lists the data catalogs in the current Amazon Web Services account.

        :param next_token: A token generated by the Athena service that specifies where to continue
        pagination if a previous request was truncated.
        :param max_results: Specifies the maximum number of data catalogs to return.
        :returns: ListDataCatalogsOutput
        :raises InternalServerException:
        :raises InvalidRequestException:
        """
        raise NotImplementedError

    @handler("ListDatabases")
    def list_databases(
        self,
        context: RequestContext,
        catalog_name: CatalogNameString,
        next_token: Token = None,
        max_results: MaxDatabasesCount = None,
    ) -> ListDatabasesOutput:
        """Lists the databases in the specified data catalog.

        :param catalog_name: The name of the data catalog that contains the databases to return.
        :param next_token: A token generated by the Athena service that specifies where to continue
        pagination if a previous request was truncated.
        :param max_results: Specifies the maximum number of results to return.
        :returns: ListDatabasesOutput
        :raises InternalServerException:
        :raises InvalidRequestException:
        :raises MetadataException:
        """
        raise NotImplementedError

    @handler("ListEngineVersions")
    def list_engine_versions(
        self,
        context: RequestContext,
        next_token: Token = None,
        max_results: MaxEngineVersionsCount = None,
    ) -> ListEngineVersionsOutput:
        """Returns a list of engine versions that are available to choose from,
        including the Auto option.

        :param next_token: A token generated by the Athena service that specifies where to continue
        pagination if a previous request was truncated.
        :param max_results: The maximum number of engine versions to return in this request.
        :returns: ListEngineVersionsOutput
        :raises InternalServerException:
        :raises InvalidRequestException:
        """
        raise NotImplementedError

    @handler("ListNamedQueries")
    def list_named_queries(
        self,
        context: RequestContext,
        next_token: Token = None,
        max_results: MaxNamedQueriesCount = None,
        work_group: WorkGroupName = None,
    ) -> ListNamedQueriesOutput:
        """Provides a list of available query IDs only for queries saved in the
        specified workgroup. Requires that you have access to the specified
        workgroup. If a workgroup is not specified, lists the saved queries for
        the primary workgroup.

        For code samples using the Amazon Web Services SDK for Java, see
        `Examples and Code
        Samples <http://docs.aws.amazon.com/athena/latest/ug/code-samples.html>`__
        in the *Amazon Athena User Guide*.

        :param next_token: A token generated by the Athena service that specifies where to continue
        pagination if a previous request was truncated.
        :param max_results: The maximum number of queries to return in this request.
        :param work_group: The name of the workgroup from which the named queries are being
        returned.
        :returns: ListNamedQueriesOutput
        :raises InternalServerException:
        :raises InvalidRequestException:
        """
        raise NotImplementedError

    @handler("ListPreparedStatements")
    def list_prepared_statements(
        self,
        context: RequestContext,
        work_group: WorkGroupName,
        next_token: Token = None,
        max_results: MaxPreparedStatementsCount = None,
    ) -> ListPreparedStatementsOutput:
        """Lists the prepared statements in the specfied workgroup.

        :param work_group: The workgroup to list the prepared statements for.
        :param next_token: A token generated by the Athena service that specifies where to continue
        pagination if a previous request was truncated.
        :param max_results: The maximum number of results to return in this request.
        :returns: ListPreparedStatementsOutput
        :raises InternalServerException:
        :raises InvalidRequestException:
        """
        raise NotImplementedError

    @handler("ListQueryExecutions")
    def list_query_executions(
        self,
        context: RequestContext,
        next_token: Token = None,
        max_results: MaxQueryExecutionsCount = None,
        work_group: WorkGroupName = None,
    ) -> ListQueryExecutionsOutput:
        """Provides a list of available query execution IDs for the queries in the
        specified workgroup. If a workgroup is not specified, returns a list of
        query execution IDs for the primary workgroup. Requires you to have
        access to the workgroup in which the queries ran.

        For code samples using the Amazon Web Services SDK for Java, see
        `Examples and Code
        Samples <http://docs.aws.amazon.com/athena/latest/ug/code-samples.html>`__
        in the *Amazon Athena User Guide*.

        :param next_token: A token generated by the Athena service that specifies where to continue
        pagination if a previous request was truncated.
        :param max_results: The maximum number of query executions to return in this request.
        :param work_group: The name of the workgroup from which queries are being returned.
        :returns: ListQueryExecutionsOutput
        :raises InternalServerException:
        :raises InvalidRequestException:
        """
        raise NotImplementedError

    @handler("ListTableMetadata")
    def list_table_metadata(
        self,
        context: RequestContext,
        catalog_name: CatalogNameString,
        database_name: NameString,
        expression: ExpressionString = None,
        next_token: Token = None,
        max_results: MaxTableMetadataCount = None,
    ) -> ListTableMetadataOutput:
        """Lists the metadata for the tables in the specified data catalog
        database.

        :param catalog_name: The name of the data catalog for which table metadata should be
        returned.
        :param database_name: The name of the database for which table metadata should be returned.
        :param expression: A regex filter that pattern-matches table names.
        :param next_token: A token generated by the Athena service that specifies where to continue
        pagination if a previous request was truncated.
        :param max_results: Specifies the maximum number of results to return.
        :returns: ListTableMetadataOutput
        :raises InternalServerException:
        :raises InvalidRequestException:
        :raises MetadataException:
        """
        raise NotImplementedError

    @handler("ListTagsForResource")
    def list_tags_for_resource(
        self,
        context: RequestContext,
        resource_arn: AmazonResourceName,
        next_token: Token = None,
        max_results: MaxTagsCount = None,
    ) -> ListTagsForResourceOutput:
        """Lists the tags associated with an Athena workgroup or data catalog
        resource.

        :param resource_arn: Lists the tags for the resource with the specified ARN.
        :param next_token: The token for the next set of results, or null if there are no
        additional results for this request, where the request lists the tags
        for the resource with the specified ARN.
        :param max_results: The maximum number of results to be returned per request that lists the
        tags for the resource.
        :returns: ListTagsForResourceOutput
        :raises InternalServerException:
        :raises InvalidRequestException:
        :raises ResourceNotFoundException:
        """
        raise NotImplementedError

    @handler("ListWorkGroups")
    def list_work_groups(
        self,
        context: RequestContext,
        next_token: Token = None,
        max_results: MaxWorkGroupsCount = None,
    ) -> ListWorkGroupsOutput:
        """Lists available workgroups for the account.

        :param next_token: A token generated by the Athena service that specifies where to continue
        pagination if a previous request was truncated.
        :param max_results: The maximum number of workgroups to return in this request.
        :returns: ListWorkGroupsOutput
        :raises InternalServerException:
        :raises InvalidRequestException:
        """
        raise NotImplementedError

    @handler("StartQueryExecution")
    def start_query_execution(
        self,
        context: RequestContext,
        query_string: QueryString,
        client_request_token: IdempotencyToken = None,
        query_execution_context: QueryExecutionContext = None,
        result_configuration: ResultConfiguration = None,
        work_group: WorkGroupName = None,
    ) -> StartQueryExecutionOutput:
        """Runs the SQL query statements contained in the ``Query``. Requires you
        to have access to the workgroup in which the query ran. Running queries
        against an external catalog requires GetDataCatalog permission to the
        catalog. For code samples using the Amazon Web Services SDK for Java,
        see `Examples and Code
        Samples <http://docs.aws.amazon.com/athena/latest/ug/code-samples.html>`__
        in the *Amazon Athena User Guide*.

        :param query_string: The SQL query statements to be executed.
        :param client_request_token: A unique case-sensitive string used to ensure the request to create the
        query is idempotent (executes only once).
        :param query_execution_context: The database within which the query executes.
        :param result_configuration: Specifies information about where and how to save the results of the
        query execution.
        :param work_group: The name of the workgroup in which the query is being started.
        :returns: StartQueryExecutionOutput
        :raises InternalServerException:
        :raises InvalidRequestException:
        :raises TooManyRequestsException:
        """
        raise NotImplementedError

    @handler("StopQueryExecution")
    def stop_query_execution(
        self, context: RequestContext, query_execution_id: QueryExecutionId
    ) -> StopQueryExecutionOutput:
        """Stops a query execution. Requires you to have access to the workgroup in
        which the query ran.

        For code samples using the Amazon Web Services SDK for Java, see
        `Examples and Code
        Samples <http://docs.aws.amazon.com/athena/latest/ug/code-samples.html>`__
        in the *Amazon Athena User Guide*.

        :param query_execution_id: The unique ID of the query execution to stop.
        :returns: StopQueryExecutionOutput
        :raises InternalServerException:
        :raises InvalidRequestException:
        """
        raise NotImplementedError

    @handler("TagResource")
    def tag_resource(
        self, context: RequestContext, resource_arn: AmazonResourceName, tags: TagList
    ) -> TagResourceOutput:
        """Adds one or more tags to an Athena resource. A tag is a label that you
        assign to a resource. In Athena, a resource can be a workgroup or data
        catalog. Each tag consists of a key and an optional value, both of which
        you define. For example, you can use tags to categorize Athena
        workgroups or data catalogs by purpose, owner, or environment. Use a
        consistent set of tag keys to make it easier to search and filter
        workgroups or data catalogs in your account. For best practices, see
        `Tagging Best
        Practices <https://aws.amazon.com/answers/account-management/aws-tagging-strategies/>`__.
        Tag keys can be from 1 to 128 UTF-8 Unicode characters, and tag values
        can be from 0 to 256 UTF-8 Unicode characters. Tags can use letters and
        numbers representable in UTF-8, and the following characters: + - = . _
        : / @. Tag keys and values are case-sensitive. Tag keys must be unique
        per resource. If you specify more than one tag, separate them by commas.

        :param resource_arn: Specifies the ARN of the Athena resource (workgroup or data catalog) to
        which tags are to be added.
        :param tags: A collection of one or more tags, separated by commas, to be added to an
        Athena workgroup or data catalog resource.
        :returns: TagResourceOutput
        :raises InternalServerException:
        :raises InvalidRequestException:
        :raises ResourceNotFoundException:
        """
        raise NotImplementedError

    @handler("UntagResource")
    def untag_resource(
        self, context: RequestContext, resource_arn: AmazonResourceName, tag_keys: TagKeyList
    ) -> UntagResourceOutput:
        """Removes one or more tags from a data catalog or workgroup resource.

        :param resource_arn: Specifies the ARN of the resource from which tags are to be removed.
        :param tag_keys: A comma-separated list of one or more tag keys whose tags are to be
        removed from the specified resource.
        :returns: UntagResourceOutput
        :raises InternalServerException:
        :raises InvalidRequestException:
        :raises ResourceNotFoundException:
        """
        raise NotImplementedError

    @handler("UpdateDataCatalog", expand=False)
    def update_data_catalog(
        self, context: RequestContext, request: UpdateDataCatalogInput
    ) -> UpdateDataCatalogOutput:
        """Updates the data catalog that has the specified name.

        :param name: The name of the data catalog to update.
        :param type: Specifies the type of data catalog to update.
        :param description: New or modified text that describes the data catalog.
        :param parameters: Specifies the Lambda function or functions to use for updating the data
        catalog.
        :returns: UpdateDataCatalogOutput
        :raises InternalServerException:
        :raises InvalidRequestException:
        """
        raise NotImplementedError

    @handler("UpdateNamedQuery")
    def update_named_query(
        self,
        context: RequestContext,
        named_query_id: NamedQueryId,
        name: NameString,
        query_string: QueryString,
        description: NamedQueryDescriptionString = None,
    ) -> UpdateNamedQueryOutput:
        """Updates a NamedQuery object. The database or workgroup cannot be
        updated.

        :param named_query_id: The unique identifier (UUID) of the query.
        :param name: The name of the query.
        :param query_string: The contents of the query with all query statements.
        :param description: The query description.
        :returns: UpdateNamedQueryOutput
        :raises InternalServerException:
        :raises InvalidRequestException:
        """
        raise NotImplementedError

    @handler("UpdatePreparedStatement")
    def update_prepared_statement(
        self,
        context: RequestContext,
        statement_name: StatementName,
        work_group: WorkGroupName,
        query_statement: QueryString,
        description: DescriptionString = None,
    ) -> UpdatePreparedStatementOutput:
        """Updates a prepared statement.

        :param statement_name: The name of the prepared statement.
        :param work_group: The workgroup for the prepared statement.
        :param query_statement: The query string for the prepared statement.
        :param description: The description of the prepared statement.
        :returns: UpdatePreparedStatementOutput
        :raises InternalServerException:
        :raises InvalidRequestException:
        :raises ResourceNotFoundException:
        """
        raise NotImplementedError

    @handler("UpdateWorkGroup")
    def update_work_group(
        self,
        context: RequestContext,
        work_group: WorkGroupName,
        description: WorkGroupDescriptionString = None,
        configuration_updates: WorkGroupConfigurationUpdates = None,
        state: WorkGroupState = None,
    ) -> UpdateWorkGroupOutput:
        """Updates the workgroup with the specified name. The workgroup's name
        cannot be changed.

        :param work_group: The specified workgroup that will be updated.
        :param description: The workgroup description.
        :param configuration_updates: The workgroup configuration that will be updated for the given
        workgroup.
        :param state: The workgroup state that will be updated for the given workgroup.
        :returns: UpdateWorkGroupOutput
        :raises InternalServerException:
        :raises InvalidRequestException:
        """
        raise NotImplementedError



from localstack_ext import __version__

# version of localstack-ext
# TODO deprecated - use localstack-ext.__version__ instead
VERSION = __version__

# default expiry seconds for Cognito access tokens
TOKEN_EXPIRY_SECONDS = 24 * 60 * 60

# name of Docker registry for Lambda images
DEFAULT_LAMBDA_DOCKER_REGISTRY = "localstack/lambda"

# request path for local pod management API
API_PATH_PODS = "/_pods"

# name of S3 bucket containing assets that are lazily downloaded - TODO: unify assets bucket layout/naming
ASSETS_S3_BUCKET = "localstack-web-assets"

# default Spark version
DEFAULT_SPARK_VERSION = "2.4.3"
# folder to persist static fields in RegionBackend and moto backend classes
STATIC_DETAILS_FOLDER_NAME = "static"
# folder to persist miscellaneous global fields used in different API implementations
EXTENDED_DETAILS_FOLDER_NAME = "extended"
REGION_STATE_FILE = "region_state"
SHARED_STATE_FILE = "shared_state"
MOTO_BACKEND_STATE_FILE = "backend_state"
UPDATE_HTTP_METHODS = ["POST", "PUT", "DELETE", "PATCH"]



_J='delete'
_I='create'
_H='ScalingTargetId'
_G='application-autoscaling'
_F='parameters'
_E='function'
_D='PolicyName'
_C='ScalableDimension'
_B='ResourceId'
_A='ServiceNamespace'
from localstack.utils.aws import aws_stack
from localstack.utils.common import select_attributes
from localstack_ext.services.cloudformation.models.servicediscovery import ServiceDiscoveryNamespace
class ApplicationAutoScalingPolicy(ServiceDiscoveryNamespace):
	@staticmethod
	def cloudformation_type():return'AWS::ApplicationAutoScaling::ScalingPolicy'
	def fetch_state(B,stack_name,resources):
		E=resources;D=stack_name;G=aws_stack.connect_to_service(_G);C=B.props;A=C.get(_A);F=C.get(_H)
		if F:A=F.split('|')[-1]
		H=B.resolve_refs_recursively(D,C[_D],E);A=B.resolve_refs_recursively(D,A,E);I=G.describe_scaling_policies(ServiceNamespace=A)['ScalingPolicies'];J=[A for A in I if A[_D]==H];return(J or[None])[0]
	def get_physical_resource_id(A,attribute,**B):return A.props.get('PolicyARN')
	@classmethod
	def get_deploy_templates(B):
		def A(params,**F):
			C=params;E=_D,_A,_B,_C,'PolicyType','StepScalingPolicyConfiguration','TargetTrackingScalingPolicyConfiguration';A=select_attributes(C,E);D=C.get(_H)
			if D:B=D.split('|');A.setdefault(_B,B[0]);A.setdefault(_C,B[1]);A.setdefault(_A,B[2])
			return A
		return{_I:{_E:'put_scaling_policy',_F:A},_J:{_E:'delete_scaling_policy',_F:[_D,_A,_B,_C]}}
class ApplicationAutoScalingScalableTarget(ServiceDiscoveryNamespace):
	@staticmethod
	def cloudformation_type():return'AWS::ApplicationAutoScaling::ScalableTarget'
	def fetch_state(A,stack_name,resources):C=resources;B=stack_name;F=aws_stack.connect_to_service(_G);D=A.props;G=A.resolve_refs_recursively(B,D[_B],C);H=A.resolve_refs_recursively(B,D[_C],C);E=A.resolve_refs_recursively(B,D[_A],C);I=F.describe_scalable_targets(ServiceNamespace=E)['ScalableTargets'];J=[A for A in I if A[_B]==G and A[_C]==H and A[_A]==E];return(J or[None])[0]
	def get_physical_resource_id(B,attribute,**C):
		if not B.state:return
		A=B.props;return'service/%s|%s|%s'%(A[_B],A[_C],A[_A])
	@classmethod
	def get_deploy_templates(A):return{_I:{_E:'register_scalable_target',_F:[_A,_B,_C,'MinCapacity','MaxCapacity','RoleARN','SuspendedState']},_J:{_E:'deregister_scalable_target',_F:[_A,_B,_C]}}


_A=False
import logging,os,re,threading,traceback
from typing import List
import localstack
from localstack import constants
from localstack.services.install import ARTIFACTS_REPO,INSTALL_DIR_STEPFUNCTIONS,INSTALL_PATH_STEPFUNCTIONS_JAR,JAR_URLS,SFN_PATCH_URL_PREFIX,Installer,InstallerRepository,add_file_to_jar
from localstack.utils.common import download,file_exists_not_empty,now,run,save_file
from localstack_ext.bootstrap.installer import mariadb_installer,mosquitto_installer,postgres_installer,redis_installer,timescaledb_installer
from localstack_ext.bootstrap.licensing import api_key_configured,is_enterprise,prepare_environment
LOG=logging.getLogger(__name__)
RULE_ENGINE_INSTALL_URL='https://github.com/whummer/serverless-iot-offline'
H2_DOWNLOAD_URL='http://www.h2database.com/h2-2019-10-14.zip'
SSL_CERT_URL=f"{ARTIFACTS_REPO}/raw/master/local-certs/server.key"
SSL_CERT_URL_FALLBACK='{api_endpoint}/proxy/localstack.cert.key'
INFRA_DIR=os.path.join(os.path.dirname(localstack.__file__),'infra')
LOCALSTACK_DIR=os.path.dirname(localstack.__file__)
POSTGRES_LIB_FOLDER='/usr/lib/postgresql/11/lib'
SFN_PATCH_PRO_CLASS1='cloud/localstack/PersistenceAspect.class'
SFN_PATCH_PRO_CLASS2='cloud/localstack/PersistenceContext.class'
SFN_PATCH_PRO_CLASS3='cloud/localstack/PersistenceState.class'
SFN_PATCH_PRO_CLASS4='cloud/localstack/PersistenceRegionState.class'
SFN_PATCH_PRO_FILE_METAINF='META-INF/aop-pro.xml'
MAVEN_REPO='https://repo1.maven.org/maven2'
URL_KRYO=f"{MAVEN_REPO}/com/esotericsoftware/kryo/5.2.0/kryo-5.2.0.jar"
URL_OBJENESIS=f"{MAVEN_REPO}/org/objenesis/objenesis/3.2/objenesis-3.2.jar"
URL_MINLOG=f"{MAVEN_REPO}/com/esotericsoftware/minlog/1.3.1/minlog-1.3.1.jar"
URL_REFLECTASM=f"{MAVEN_REPO}/com/esotericsoftware/reflectasm/1.11.9/reflectasm-1.11.9.jar"
PRO_JAR_URLS=[URL_KRYO,URL_OBJENESIS,URL_MINLOG,URL_REFLECTASM]
INSTALL_LOCK=threading.RLock()
def install_libs():install_iot_rule_engine();postgres_installer.install(raise_on_error=_A);timescaledb_installer.install(raise_on_error=_A);redis_installer.install(raise_on_error=_A);mosquitto_installer.install(raise_on_error=_A);install_stepfunctions()
def install_iot_rule_engine():
	A=LOCALSTACK_DIR;B=os.path.join(A,'node_modules','serverless-iot-offline','query.js')
	if not os.path.exists(B):LOG.info('Installing IoT rule engine. This may take a while.');run('cd %s; npm install %s'%(A,RULE_ENGINE_INSTALL_URL))
	return B
def install_stepfunctions():
	E='META-INF/MANIFEST.MF'
	if not os.path.exists(INSTALL_PATH_STEPFUNCTIONS_JAR):LOG.warning('StepFunctions JAR not available - skipping installation');return
	F=[SFN_PATCH_PRO_CLASS1,SFN_PATCH_PRO_CLASS2,SFN_PATCH_PRO_CLASS3,SFN_PATCH_PRO_CLASS4,SFN_PATCH_PRO_FILE_METAINF]
	for B in F:G=f"{SFN_PATCH_URL_PREFIX}/{B}";add_file_to_jar(B,G,target_jar=INSTALL_PATH_STEPFUNCTIONS_JAR)
	H=os.path.join(INSTALL_DIR_STEPFUNCTIONS,'META-INF','MANIFEST.MF');A=run(['unzip','-p',INSTALL_PATH_STEPFUNCTIONS_JAR,E]);A=re.sub('Main-Class: .+','Main-Class: cloud.localstack.StepFunctionsStarter',A);I=' '.join([os.path.basename(A)for A in[*(PRO_JAR_URLS),*(JAR_URLS)]]);A=re.sub('Class-Path: (.+ )\\. ',f"Class-Path: {I} . ",A);save_file(H,A);run(['zip',INSTALL_PATH_STEPFUNCTIONS_JAR,E],cwd=INSTALL_DIR_STEPFUNCTIONS)
	for C in PRO_JAR_URLS:
		D=os.path.join(INSTALL_DIR_STEPFUNCTIONS,os.path.basename(C))
		if not file_exists_not_empty(D):download(C,D)
def setup_ssl_cert():
	from localstack.services import generic_proxy as C;A=C.get_cert_pem_file_path()
	if os.path.exists(A):
		if is_enterprise():LOG.debug('Avoiding to update SSL certificate.');return
		D=6*60*60;E=os.path.getmtime(A)
		if E>now()-D:LOG.debug('Using cached SSL certificate (less than 6hrs since last update).');return
	LOG.debug('Attempting to download local SSL certificate file');F=3;G=5
	try:return download_github_artifact(SSL_CERT_URL,A,timeout=F)
	except Exception:
		B=SSL_CERT_URL_FALLBACK.format(api_endpoint=constants.API_ENDPOINT)
		try:return download(B,A,timeout=G)
		except Exception as H:LOG.info('Unable to download local test SSL certificate from %s to %s (using self-signed cert as fallback): %s',B,A,H);raise
def download_github_artifact(url,target_file,timeout=None):
	B=target_file;A=url
	def C(url,print_error=_A):
		try:download(url,B,timeout=timeout);return True
		except Exception as A:
			if print_error:LOG.info('Unable to download Github artifact from from %s to %s: %s %s'%(url,B,A,traceback.format_exc()))
	D=C(A)
	if not D:A=A.replace('https://github.com','https://cdn.jsdelivr.net/gh');A=A.replace('/raw/master/','@master/');C(A,True)
def install_azure():from localstack_ext.services.azure import api_specs as A;A.download_api_specs()
def install_ecr():from localstack_ext.services.ecr import registry as A;A.get_registry_binary()
def install_pysiddhi():from localstack_ext.services.kinesisanalytics import query_utils as A;A.setup_siddhi()
def install_neptune():from localstack_ext.services.neptune import neptune_api as A;A.install_graphdb()
def install_qldb():from localstack_ext.services.qldb import partiql as A;A.install_partiql()
def install_k3d():from localstack_ext.services.eks import k8s_utils as A;A.KubeProviderK3S().initialize()
def install_kafka():from localstack_ext.services.kafka.provider import download_kafka as A;A()
class ExtInstallerRepository(InstallerRepository):
	name='ext'
	def should_load(A):return api_key_configured()
	def load(A,*B,**C):
		LOG.debug('Preparing Pro environment for LocalStack Package Manager.')
		with prepare_environment():LOG.debug('Pro environment has successfully been prepared.')
	def get_installer(A):return[('iot-rule-engine',install_iot_rule_engine),('postgres',postgres_installer.install),('timescaledb',timescaledb_installer.install),('redis',redis_installer.install),('mqtt',mosquitto_installer.install),('azure',install_azure),('ecr',install_ecr),('mysql',mariadb_installer.install),('pysiddhi',install_pysiddhi),('neptune',install_neptune),('qldb',install_qldb),('k3d',install_k3d),('kafka',install_kafka),('stepfunctions',install_stepfunctions)]


_H='parameters'
_G='delete'
_F='create'
_E='timestream-write'
_D='Properties'
_C='function'
_B='TableName'
_A='DatabaseName'
from typing import Dict
from localstack.services.cloudformation.deployment_utils import generate_default_name_without_stack
from localstack.services.cloudformation.service_models import GenericBaseModel
from localstack.utils.aws import aws_stack
from localstack.utils.common import short_uid
class TimestreamDatabase(GenericBaseModel):
	@staticmethod
	def cloudformation_type():return'AWS::Timestream::Database'
	@staticmethod
	def add_defaults(resource,stack_name):
		A=resource;B=A[_D]
		if not B.get(_A):B[_A]=generate_default_name_without_stack(A['LogicalResourceId'])
	def fetch_state(A,stack_name,resources):B=aws_stack.connect_to_service(_E);C=A.resolve_refs_recursively(stack_name,A.props.get(_A),resources);D=B.describe_database(DatabaseName=C);return D['Database']
	def get_physical_resource_id(A,attribute,**B):return A.props.get(_A)
	@staticmethod
	def get_deploy_templates():return{_F:{_C:'create_database'},_G:{_C:'delete_database',_H:[_A]}}
class TimestreamTable(GenericBaseModel):
	@staticmethod
	def cloudformation_type():return'AWS::Timestream::Table'
	@staticmethod
	def add_defaults(resource,stack_name):
		A=resource[_D]
		if not A.get(_B):A[_B]=f"t{short_uid()}"
	def fetch_state(A,stack_name,resources):
		C=resources;B=stack_name;E=aws_stack.connect_to_service(_E);D=A.resolve_refs_recursively(B,A.props.get(_A),C);F=A.resolve_refs_recursively(B,A.props.get(_B),C)
		if not D:return
		G=E.describe_table(DatabaseName=D,TableName=F);return G['Table']
	def get_physical_resource_id(B,attribute,**C):A=B.props;return f"{A.get(_A)}|{A.get(_B)}"
	@staticmethod
	def get_deploy_templates():return{_F:{_C:'create_table'},_G:{_C:'delete_table',_H:[_A,_B]}}


import sys
from datetime import datetime
from typing import Dict, List, Optional

if sys.version_info >= (3, 8):
    from typing import TypedDict
else:
    from typing_extensions import TypedDict

from localstack.aws.api import RequestContext, ServiceException, ServiceRequest, handler

AirflowVersion = str
CloudWatchLogGroupArn = str
ConfigKey = str
ConfigValue = str
Double = float
EnvironmentArn = str
EnvironmentClass = str
EnvironmentName = str
ErrorCode = str
ErrorMessage = str
Hostname = str
IamRoleArn = str
Integer = int
KmsKey = str
ListEnvironmentsInputMaxResultsInteger = int
LoggingEnabled = bool
MaxWorkers = int
MinWorkers = int
NextToken = str
RelativePath = str
S3BucketArn = str
S3ObjectVersion = str
Schedulers = int
SecurityGroupId = str
String = str
SubnetId = str
SyntheticCreateCliTokenResponseToken = str
SyntheticCreateWebLoginTokenResponseToken = str
TagKey = str
TagValue = str
UpdateSource = str
WebserverUrl = str
WeeklyMaintenanceWindowStart = str


class EnvironmentStatus(str):
    CREATING = "CREATING"
    CREATE_FAILED = "CREATE_FAILED"
    AVAILABLE = "AVAILABLE"
    UPDATING = "UPDATING"
    DELETING = "DELETING"
    DELETED = "DELETED"
    UNAVAILABLE = "UNAVAILABLE"
    UPDATE_FAILED = "UPDATE_FAILED"


class LoggingLevel(str):
    CRITICAL = "CRITICAL"
    ERROR = "ERROR"
    WARNING = "WARNING"
    INFO = "INFO"
    DEBUG = "DEBUG"


class Unit(str):
    Seconds = "Seconds"
    Microseconds = "Microseconds"
    Milliseconds = "Milliseconds"
    Bytes = "Bytes"
    Kilobytes = "Kilobytes"
    Megabytes = "Megabytes"
    Gigabytes = "Gigabytes"
    Terabytes = "Terabytes"
    Bits = "Bits"
    Kilobits = "Kilobits"
    Megabits = "Megabits"
    Gigabits = "Gigabits"
    Terabits = "Terabits"
    Percent = "Percent"
    Count = "Count"
    Bytes_Second = "Bytes/Second"
    Kilobytes_Second = "Kilobytes/Second"
    Megabytes_Second = "Megabytes/Second"
    Gigabytes_Second = "Gigabytes/Second"
    Terabytes_Second = "Terabytes/Second"
    Bits_Second = "Bits/Second"
    Kilobits_Second = "Kilobits/Second"
    Megabits_Second = "Megabits/Second"
    Gigabits_Second = "Gigabits/Second"
    Terabits_Second = "Terabits/Second"
    Count_Second = "Count/Second"
    None_ = "None"


class UpdateStatus(str):
    SUCCESS = "SUCCESS"
    PENDING = "PENDING"
    FAILED = "FAILED"


class WebserverAccessMode(str):
    PRIVATE_ONLY = "PRIVATE_ONLY"
    PUBLIC_ONLY = "PUBLIC_ONLY"


class AccessDeniedException(ServiceException):
    """Access to the Apache Airflow Web UI or CLI has been denied due to
    insufficient permissions. To learn more, see `Accessing an Amazon MWAA
    environment <https://docs.aws.amazon.com/mwaa/latest/userguide/access-policies.html>`__.
    """

    Message: Optional[String]


class InternalServerException(ServiceException):
    """InternalServerException: An internal error has occurred."""

    message: Optional[String]


class ResourceNotFoundException(ServiceException):
    """ResourceNotFoundException: The resource is not available."""

    message: Optional[String]


class ValidationException(ServiceException):
    """ValidationException: The provided input is not valid."""

    message: Optional[String]


AirflowConfigurationOptions = Dict[ConfigKey, ConfigValue]


class CreateCliTokenRequest(ServiceRequest):
    Name: EnvironmentName


class CreateCliTokenResponse(TypedDict, total=False):
    CliToken: Optional[SyntheticCreateCliTokenResponseToken]
    WebServerHostname: Optional[Hostname]


TagMap = Dict[TagKey, TagValue]
SubnetList = List[SubnetId]
SecurityGroupList = List[SecurityGroupId]


class NetworkConfiguration(TypedDict, total=False):
    """Describes the VPC networking components used to secure and enable
    network traffic between the Amazon Web Services resources for your
    environment. To learn more, see `About networking on Amazon
    MWAA <https://docs.aws.amazon.com/mwaa/latest/userguide/networking-about.html>`__.
    """

    SecurityGroupIds: Optional[SecurityGroupList]
    SubnetIds: Optional[SubnetList]


class ModuleLoggingConfigurationInput(TypedDict, total=False):
    """Enables the Apache Airflow log type (e.g. ``DagProcessingLogs``) and
    defines the log level to send to CloudWatch Logs (e.g. ``INFO``).
    """

    Enabled: LoggingEnabled
    LogLevel: LoggingLevel


class LoggingConfigurationInput(TypedDict, total=False):
    """Defines the Apache Airflow log types to send to CloudWatch Logs."""

    DagProcessingLogs: Optional[ModuleLoggingConfigurationInput]
    SchedulerLogs: Optional[ModuleLoggingConfigurationInput]
    TaskLogs: Optional[ModuleLoggingConfigurationInput]
    WebserverLogs: Optional[ModuleLoggingConfigurationInput]
    WorkerLogs: Optional[ModuleLoggingConfigurationInput]


SyntheticCreateEnvironmentInputAirflowConfigurationOptions = Dict[ConfigKey, ConfigValue]


class CreateEnvironmentInput(ServiceRequest):
    """This section contains the Amazon Managed Workflows for Apache Airflow
    (MWAA) API reference documentation to create an environment. For more
    information, see `Get started with Amazon Managed Workflows for Apache
    Airflow <https://docs.aws.amazon.com/mwaa/latest/userguide/get-started.html>`__.
    """

    AirflowConfigurationOptions: Optional[
        SyntheticCreateEnvironmentInputAirflowConfigurationOptions
    ]
    AirflowVersion: Optional[AirflowVersion]
    DagS3Path: RelativePath
    EnvironmentClass: Optional[EnvironmentClass]
    ExecutionRoleArn: IamRoleArn
    KmsKey: Optional[KmsKey]
    LoggingConfiguration: Optional[LoggingConfigurationInput]
    MaxWorkers: Optional[MaxWorkers]
    MinWorkers: Optional[MinWorkers]
    Name: EnvironmentName
    NetworkConfiguration: NetworkConfiguration
    PluginsS3ObjectVersion: Optional[S3ObjectVersion]
    PluginsS3Path: Optional[RelativePath]
    RequirementsS3ObjectVersion: Optional[S3ObjectVersion]
    RequirementsS3Path: Optional[RelativePath]
    Schedulers: Optional[Schedulers]
    SourceBucketArn: S3BucketArn
    Tags: Optional[TagMap]
    WebserverAccessMode: Optional[WebserverAccessMode]
    WeeklyMaintenanceWindowStart: Optional[WeeklyMaintenanceWindowStart]


class CreateEnvironmentOutput(TypedDict, total=False):
    Arn: Optional[EnvironmentArn]


class CreateWebLoginTokenRequest(ServiceRequest):
    Name: EnvironmentName


class CreateWebLoginTokenResponse(TypedDict, total=False):
    WebServerHostname: Optional[Hostname]
    WebToken: Optional[SyntheticCreateWebLoginTokenResponseToken]


CreatedAt = datetime


class DeleteEnvironmentInput(ServiceRequest):
    Name: EnvironmentName


class DeleteEnvironmentOutput(TypedDict, total=False):
    pass


class Dimension(TypedDict, total=False):
    """**Internal only**. Represents the dimensions of a metric. To learn more
    about the metrics published to Amazon CloudWatch, see `Amazon MWAA
    performance metrics in Amazon
    CloudWatch <https://docs.aws.amazon.com/mwaa/latest/userguide/cw-metrics.html>`__.
    """

    Name: String
    Value: String


Dimensions = List[Dimension]


class ModuleLoggingConfiguration(TypedDict, total=False):
    """Describes the Apache Airflow log details for the log type (e.g.
    ``DagProcessingLogs``).
    """

    CloudWatchLogGroupArn: Optional[CloudWatchLogGroupArn]
    Enabled: Optional[LoggingEnabled]
    LogLevel: Optional[LoggingLevel]


class LoggingConfiguration(TypedDict, total=False):
    """Describes the Apache Airflow log types that are published to CloudWatch
    Logs.
    """

    DagProcessingLogs: Optional[ModuleLoggingConfiguration]
    SchedulerLogs: Optional[ModuleLoggingConfiguration]
    TaskLogs: Optional[ModuleLoggingConfiguration]
    WebserverLogs: Optional[ModuleLoggingConfiguration]
    WorkerLogs: Optional[ModuleLoggingConfiguration]


class UpdateError(TypedDict, total=False):
    """Describes the error(s) encountered with the last update of the
    environment.
    """

    ErrorCode: Optional[ErrorCode]
    ErrorMessage: Optional[ErrorMessage]


UpdateCreatedAt = datetime


class LastUpdate(TypedDict, total=False):
    """Describes the status of the last update on the environment, and any
    errors that were encountered.
    """

    CreatedAt: Optional[UpdateCreatedAt]
    Error: Optional[UpdateError]
    Source: Optional[UpdateSource]
    Status: Optional[UpdateStatus]


class Environment(TypedDict, total=False):
    """Describes an Amazon Managed Workflows for Apache Airflow (MWAA)
    environment.
    """

    AirflowConfigurationOptions: Optional[AirflowConfigurationOptions]
    AirflowVersion: Optional[AirflowVersion]
    Arn: Optional[EnvironmentArn]
    CreatedAt: Optional[CreatedAt]
    DagS3Path: Optional[RelativePath]
    EnvironmentClass: Optional[EnvironmentClass]
    ExecutionRoleArn: Optional[IamRoleArn]
    KmsKey: Optional[KmsKey]
    LastUpdate: Optional[LastUpdate]
    LoggingConfiguration: Optional[LoggingConfiguration]
    MaxWorkers: Optional[MaxWorkers]
    MinWorkers: Optional[MinWorkers]
    Name: Optional[EnvironmentName]
    NetworkConfiguration: Optional[NetworkConfiguration]
    PluginsS3ObjectVersion: Optional[S3ObjectVersion]
    PluginsS3Path: Optional[RelativePath]
    RequirementsS3ObjectVersion: Optional[S3ObjectVersion]
    RequirementsS3Path: Optional[RelativePath]
    Schedulers: Optional[Schedulers]
    ServiceRoleArn: Optional[IamRoleArn]
    SourceBucketArn: Optional[S3BucketArn]
    Status: Optional[EnvironmentStatus]
    Tags: Optional[TagMap]
    WebserverAccessMode: Optional[WebserverAccessMode]
    WebserverUrl: Optional[WebserverUrl]
    WeeklyMaintenanceWindowStart: Optional[WeeklyMaintenanceWindowStart]


EnvironmentList = List[EnvironmentName]


class GetEnvironmentInput(ServiceRequest):
    Name: EnvironmentName


class GetEnvironmentOutput(TypedDict, total=False):
    Environment: Optional[Environment]


class ListEnvironmentsInput(ServiceRequest):
    MaxResults: Optional[ListEnvironmentsInputMaxResultsInteger]
    NextToken: Optional[NextToken]


class ListEnvironmentsOutput(TypedDict, total=False):
    Environments: EnvironmentList
    NextToken: Optional[NextToken]


class ListTagsForResourceInput(ServiceRequest):
    ResourceArn: EnvironmentArn


class ListTagsForResourceOutput(TypedDict, total=False):
    Tags: Optional[TagMap]


Timestamp = datetime


class StatisticSet(TypedDict, total=False):
    """**Internal only**. Represents a set of statistics that describe a
    specific metric. To learn more about the metrics published to Amazon
    CloudWatch, see `Amazon MWAA performance metrics in Amazon
    CloudWatch <https://docs.aws.amazon.com/mwaa/latest/userguide/cw-metrics.html>`__.
    """

    Maximum: Optional[Double]
    Minimum: Optional[Double]
    SampleCount: Optional[Integer]
    Sum: Optional[Double]


class MetricDatum(TypedDict, total=False):
    """**Internal only**. Collects Apache Airflow metrics. To learn more about
    the metrics published to Amazon CloudWatch, see `Amazon MWAA performance
    metrics in Amazon
    CloudWatch <https://docs.aws.amazon.com/mwaa/latest/userguide/cw-metrics.html>`__.
    """

    Dimensions: Optional[Dimensions]
    MetricName: String
    StatisticValues: Optional[StatisticSet]
    Timestamp: Timestamp
    Unit: Optional[Unit]
    Value: Optional[Double]


MetricData = List[MetricDatum]


class PublishMetricsInput(ServiceRequest):
    EnvironmentName: EnvironmentName
    MetricData: MetricData


class PublishMetricsOutput(TypedDict, total=False):
    pass


SyntheticUpdateEnvironmentInputAirflowConfigurationOptions = Dict[ConfigKey, ConfigValue]
TagKeyList = List[TagKey]


class TagResourceInput(ServiceRequest):
    ResourceArn: EnvironmentArn
    Tags: TagMap


class TagResourceOutput(TypedDict, total=False):
    pass


class UntagResourceInput(ServiceRequest):
    ResourceArn: EnvironmentArn
    tagKeys: TagKeyList


class UntagResourceOutput(TypedDict, total=False):
    pass


class UpdateNetworkConfigurationInput(TypedDict, total=False):
    """Defines the VPC networking components used to secure and enable network
    traffic between the Amazon Web Services resources for your environment.
    To learn more, see `About networking on Amazon
    MWAA <https://docs.aws.amazon.com/mwaa/latest/userguide/networking-about.html>`__.
    """

    SecurityGroupIds: SecurityGroupList


class UpdateEnvironmentInput(ServiceRequest):
    AirflowConfigurationOptions: Optional[
        SyntheticUpdateEnvironmentInputAirflowConfigurationOptions
    ]
    AirflowVersion: Optional[AirflowVersion]
    DagS3Path: Optional[RelativePath]
    EnvironmentClass: Optional[EnvironmentClass]
    ExecutionRoleArn: Optional[IamRoleArn]
    LoggingConfiguration: Optional[LoggingConfigurationInput]
    MaxWorkers: Optional[MaxWorkers]
    MinWorkers: Optional[MinWorkers]
    Name: EnvironmentName
    NetworkConfiguration: Optional[UpdateNetworkConfigurationInput]
    PluginsS3ObjectVersion: Optional[S3ObjectVersion]
    PluginsS3Path: Optional[RelativePath]
    RequirementsS3ObjectVersion: Optional[S3ObjectVersion]
    RequirementsS3Path: Optional[RelativePath]
    Schedulers: Optional[Schedulers]
    SourceBucketArn: Optional[S3BucketArn]
    WebserverAccessMode: Optional[WebserverAccessMode]
    WeeklyMaintenanceWindowStart: Optional[WeeklyMaintenanceWindowStart]


class UpdateEnvironmentOutput(TypedDict, total=False):
    Arn: Optional[EnvironmentArn]


class MwaaApi:

    service = "mwaa"
    version = "2020-07-01"

    @handler("CreateCliToken")
    def create_cli_token(
        self, context: RequestContext, name: EnvironmentName
    ) -> CreateCliTokenResponse:
        """Creates a CLI token for the Airflow CLI. To learn more, see `Creating an
        Apache Airflow CLI
        token <https://docs.aws.amazon.com/mwaa/latest/userguide/call-mwaa-apis-cli.html>`__.

        :param name: The name of the Amazon MWAA environment.
        :returns: CreateCliTokenResponse
        :raises ResourceNotFoundException:
        """
        raise NotImplementedError

    @handler("CreateEnvironment")
    def create_environment(
        self,
        context: RequestContext,
        dag_s3_path: RelativePath,
        execution_role_arn: IamRoleArn,
        name: EnvironmentName,
        network_configuration: NetworkConfiguration,
        source_bucket_arn: S3BucketArn,
        airflow_configuration_options: SyntheticCreateEnvironmentInputAirflowConfigurationOptions = None,
        airflow_version: AirflowVersion = None,
        environment_class: EnvironmentClass = None,
        kms_key: KmsKey = None,
        logging_configuration: LoggingConfigurationInput = None,
        max_workers: MaxWorkers = None,
        min_workers: MinWorkers = None,
        plugins_s3_object_version: S3ObjectVersion = None,
        plugins_s3_path: RelativePath = None,
        requirements_s3_object_version: S3ObjectVersion = None,
        requirements_s3_path: RelativePath = None,
        schedulers: Schedulers = None,
        tags: TagMap = None,
        webserver_access_mode: WebserverAccessMode = None,
        weekly_maintenance_window_start: WeeklyMaintenanceWindowStart = None,
    ) -> CreateEnvironmentOutput:
        """Creates an Amazon Managed Workflows for Apache Airflow (MWAA)
        environment.

        :param dag_s3_path: The relative path to the DAGs folder on your Amazon S3 bucket.
        :param execution_role_arn: The Amazon Resource Name (ARN) of the execution role for your
        environment.
        :param name: The name of the Amazon MWAA environment.
        :param network_configuration: The VPC networking components used to secure and enable network traffic
        between the Amazon Web Services resources for your environment.
        :param source_bucket_arn: The Amazon Resource Name (ARN) of the Amazon S3 bucket where your DAG
        code and supporting files are stored.
        :param airflow_configuration_options: A list of key-value pairs containing the Apache Airflow configuration
        options you want to attach to your environment.
        :param airflow_version: The Apache Airflow version for your environment.
        :param environment_class: The environment class type.
        :param kms_key: The Amazon Web Services Key Management Service (KMS) key to encrypt the
        data in your environment.
        :param logging_configuration: Defines the Apache Airflow logs to send to CloudWatch Logs.
        :param max_workers: The maximum number of workers that you want to run in your environment.
        :param min_workers: The minimum number of workers that you want to run in your environment.
        :param plugins_s3_object_version: The version of the plugins.
        :param plugins_s3_path: The relative path to the ``plugins.
        :param requirements_s3_object_version: The version of the requirements.
        :param requirements_s3_path: The relative path to the ``requirements.
        :param schedulers: The number of Apache Airflow schedulers to run in your environment.
        :param tags: The key-value tag pairs you want to associate to your environment.
        :param webserver_access_mode: The Apache Airflow *Web server* access mode.
        :param weekly_maintenance_window_start: The day and time of the week in Coordinated Universal Time (UTC) 24-hour
        standard time to start weekly maintenance updates of your environment in
        the following format: ``DAY:HH:MM``.
        :returns: CreateEnvironmentOutput
        :raises ValidationException:
        :raises InternalServerException:
        """
        raise NotImplementedError

    @handler("CreateWebLoginToken")
    def create_web_login_token(
        self, context: RequestContext, name: EnvironmentName
    ) -> CreateWebLoginTokenResponse:
        """Creates a web login token for the Airflow Web UI. To learn more, see
        `Creating an Apache Airflow web login
        token <https://docs.aws.amazon.com/mwaa/latest/userguide/call-mwaa-apis-web.html>`__.

        :param name: The name of the Amazon MWAA environment.
        :returns: CreateWebLoginTokenResponse
        :raises AccessDeniedException:
        :raises ResourceNotFoundException:
        :raises ValidationException:
        :raises InternalServerException:
        """
        raise NotImplementedError

    @handler("DeleteEnvironment")
    def delete_environment(
        self, context: RequestContext, name: EnvironmentName
    ) -> DeleteEnvironmentOutput:
        """Deletes an Amazon Managed Workflows for Apache Airflow (MWAA)
        environment.

        :param name: The name of the Amazon MWAA environment.
        :returns: DeleteEnvironmentOutput
        :raises ResourceNotFoundException:
        :raises ValidationException:
        :raises InternalServerException:
        """
        raise NotImplementedError

    @handler("GetEnvironment")
    def get_environment(
        self, context: RequestContext, name: EnvironmentName
    ) -> GetEnvironmentOutput:
        """Describes an Amazon Managed Workflows for Apache Airflow (MWAA)
        environment.

        :param name: The name of the Amazon MWAA environment.
        :returns: GetEnvironmentOutput
        :raises ResourceNotFoundException:
        :raises ValidationException:
        :raises InternalServerException:
        """
        raise NotImplementedError

    @handler("ListEnvironments")
    def list_environments(
        self,
        context: RequestContext,
        max_results: ListEnvironmentsInputMaxResultsInteger = None,
        next_token: NextToken = None,
    ) -> ListEnvironmentsOutput:
        """Lists the Amazon Managed Workflows for Apache Airflow (MWAA)
        environments.

        :param max_results: The maximum number of results to retrieve per page.
        :param next_token: Retrieves the next page of the results.
        :returns: ListEnvironmentsOutput
        :raises ValidationException:
        :raises InternalServerException:
        """
        raise NotImplementedError

    @handler("ListTagsForResource")
    def list_tags_for_resource(
        self, context: RequestContext, resource_arn: EnvironmentArn
    ) -> ListTagsForResourceOutput:
        """Lists the key-value tag pairs associated to the Amazon Managed Workflows
        for Apache Airflow (MWAA) environment. For example,
        ``"Environment": "Staging"``.

        :param resource_arn: The Amazon Resource Name (ARN) of the Amazon MWAA environment.
        :returns: ListTagsForResourceOutput
        :raises ResourceNotFoundException:
        :raises ValidationException:
        :raises InternalServerException:
        """
        raise NotImplementedError

    @handler("PublishMetrics")
    def publish_metrics(
        self, context: RequestContext, environment_name: EnvironmentName, metric_data: MetricData
    ) -> PublishMetricsOutput:
        """**Internal only**. Publishes environment health metrics to Amazon
        CloudWatch.

        :param environment_name: **Internal only**.
        :param metric_data: **Internal only**.
        :returns: PublishMetricsOutput
        :raises ValidationException:
        :raises InternalServerException:
        """
        raise NotImplementedError

    @handler("TagResource")
    def tag_resource(
        self, context: RequestContext, resource_arn: EnvironmentArn, tags: TagMap
    ) -> TagResourceOutput:
        """Associates key-value tag pairs to your Amazon Managed Workflows for
        Apache Airflow (MWAA) environment.

        :param resource_arn: The Amazon Resource Name (ARN) of the Amazon MWAA environment.
        :param tags: The key-value tag pairs you want to associate to your environment.
        :returns: TagResourceOutput
        :raises ResourceNotFoundException:
        :raises ValidationException:
        :raises InternalServerException:
        """
        raise NotImplementedError

    @handler("UntagResource")
    def untag_resource(
        self, context: RequestContext, resource_arn: EnvironmentArn, tag_keys: TagKeyList
    ) -> UntagResourceOutput:
        """Removes key-value tag pairs associated to your Amazon Managed Workflows
        for Apache Airflow (MWAA) environment. For example,
        ``"Environment": "Staging"``.

        :param resource_arn: The Amazon Resource Name (ARN) of the Amazon MWAA environment.
        :param tag_keys: The key-value tag pair you want to remove.
        :returns: UntagResourceOutput
        :raises ResourceNotFoundException:
        :raises ValidationException:
        :raises InternalServerException:
        """
        raise NotImplementedError

    @handler("UpdateEnvironment")
    def update_environment(
        self,
        context: RequestContext,
        name: EnvironmentName,
        airflow_configuration_options: SyntheticUpdateEnvironmentInputAirflowConfigurationOptions = None,
        airflow_version: AirflowVersion = None,
        dag_s3_path: RelativePath = None,
        environment_class: EnvironmentClass = None,
        execution_role_arn: IamRoleArn = None,
        logging_configuration: LoggingConfigurationInput = None,
        max_workers: MaxWorkers = None,
        min_workers: MinWorkers = None,
        network_configuration: UpdateNetworkConfigurationInput = None,
        plugins_s3_object_version: S3ObjectVersion = None,
        plugins_s3_path: RelativePath = None,
        requirements_s3_object_version: S3ObjectVersion = None,
        requirements_s3_path: RelativePath = None,
        schedulers: Schedulers = None,
        source_bucket_arn: S3BucketArn = None,
        webserver_access_mode: WebserverAccessMode = None,
        weekly_maintenance_window_start: WeeklyMaintenanceWindowStart = None,
    ) -> UpdateEnvironmentOutput:
        """Updates an Amazon Managed Workflows for Apache Airflow (MWAA)
        environment.

        :param name: The name of your Amazon MWAA environment.
        :param airflow_configuration_options: A list of key-value pairs containing the Apache Airflow configuration
        options you want to attach to your environment.
        :param airflow_version: The Apache Airflow version for your environment.
        :param dag_s3_path: The relative path to the DAGs folder on your Amazon S3 bucket.
        :param environment_class: The environment class type.
        :param execution_role_arn: The Amazon Resource Name (ARN) of the execution role in IAM that allows
        MWAA to access Amazon Web Services resources in your environment.
        :param logging_configuration: The Apache Airflow log types to send to CloudWatch Logs.
        :param max_workers: The maximum number of workers that you want to run in your environment.
        :param min_workers: The minimum number of workers that you want to run in your environment.
        :param network_configuration: The VPC networking components used to secure and enable network traffic
        between the Amazon Web Services resources for your environment.
        :param plugins_s3_object_version: The version of the plugins.
        :param plugins_s3_path: The relative path to the ``plugins.
        :param requirements_s3_object_version: The version of the requirements.
        :param requirements_s3_path: The relative path to the ``requirements.
        :param schedulers: The number of Apache Airflow schedulers to run in your Amazon MWAA
        environment.
        :param source_bucket_arn: The Amazon Resource Name (ARN) of the Amazon S3 bucket where your DAG
        code and supporting files are stored.
        :param webserver_access_mode: The Apache Airflow *Web server* access mode.
        :param weekly_maintenance_window_start: The day and time of the week in Coordinated Universal Time (UTC) 24-hour
        standard time to start weekly maintenance updates of your environment in
        the following format: ``DAY:HH:MM``.
        :returns: UpdateEnvironmentOutput
        :raises ResourceNotFoundException:
        :raises ValidationException:
        :raises InternalServerException:
        """
        raise NotImplementedError



_A='Name'
from localstack.services.cloudformation.service_models import REF_ID_ATTRS,GenericBaseModel
from localstack.utils.aws import aws_stack
from localstack.utils.common import select_attributes,short_uid
def canonicalize_name(name):
	A=name
	if A[-1]!='.':return f"{A}."
	return A
class Route53HostedZone(GenericBaseModel):
	@staticmethod
	def cloudformation_type():return'AWS::Route53::HostedZone'
	def fetch_state(B,stack_name,resources):C=aws_stack.connect_to_service('route53');A=C.list_hosted_zones()['HostedZones'];D=canonicalize_name(B.props.get(_A));A=[B for B in A if B[_A]==D];return(A or[None])[0]
	def get_physical_resource_id(A,attribute=None,**B):return A.props.get('Id')
	def get_cfn_attribute(A,attribute_name):
		B=attribute_name
		if B=='NameServers':return[A.props.get(_A)]
		if B in REF_ID_ATTRS:return A.get_physical_resource_id()
		return super(Route53HostedZone,A).get_cfn_attribute(B)
	@staticmethod
	def get_deploy_templates():
		B='parameters';A='function'
		def C(resource_props,resources,resource_id,*E,**F):D=Route53HostedZone(resources[resource_id]);B=D.props;A=select_attributes(B,['HostedZoneConfig',_A]);C=B.get('VPCs',[]);A['VPC']=C and C[0]or{};A['CallerReference']=short_uid();return A
		def D(resource_props,resources,resource_id,*B,**C):A=Route53HostedZone(resources[resource_id]);return A.props.get('Id')
		return{'create':{A:'create_hosted_zone',B:C},'delete':{A:'delete_hosted_zone',B:{'Id':D}}}


import io,os,zipfile
from localstack.utils.files import mkdir
from localstack_ext.bootstrap.pods.service_state import BackendState,ServiceKey,ServiceState
def get_path_for_backend(temporary_path,service_key):A=os.path.join(temporary_path,*(service_key));mkdir(A);return A
class ServiceStateMarshaller:
	@staticmethod
	def marshal(state):
		C=state;A=io.BytesIO()
		with zipfile.ZipFile(A,'a')as D:
			for (B,G) in C.state.items():
				H=os.path.join(B.account_id,B.service,B.region)
				for (I,J) in G.backends.items():D.writestr(os.path.join('api_states',H,I),J)
			for (K,L) in C.assets.items():
				E=zipfile.ZipFile(io.BytesIO(L))
				for F in E.namelist():D.writestr(os.path.join(K,F),E.read(F))
		A.seek(0);return A.getvalue()
	@staticmethod
	def unmarshall(zip_content):
		B=zipfile.ZipFile(io.BytesIO(zip_content));C=ServiceState()
		def D():D=A.split('/');E,F,G,H=D[-4:];I=BackendState(key=ServiceKey(E,G,F),backends={H:B.read(A)});C.add(I)
		def E():D=A.split('/')[0];C.add_asset(service=D,content=B.read(A))
		for A in B.namelist():
			if A.startswith('api_'):D()
			else:E()
		return C